import warnings
import os

from rich import print
import healpy as hp
import numpy as np
import astropy.units as u
import matplotlib.pyplot as plt
import matplotlib.colors as colors
from .plottools import *
from .temp_newvisufunc import projview as temp_projview



# Fix for macos openMP duplicate bug
os.environ["KMP_DUPLICATE_LIB_OK"] = "True"


@u.quantity_input(freq=u.Hz, fwhm=(u.arcmin, u.rad, u.deg))
def plot(
    input,
    *,
    sig=0,
    comp=None,
    freq=None,
    ticks=None,
    min=None,
    max=None,
    rng=None,
    cbar=True,
    unit=None,
    fwhm=0.0 * u.arcmin,
    nside=None,
    sample=-1,
    mask=None,
    maskfill=None,
    cmap=None,
    norm=None,
    norm_dict=None,
    remove_dip=False,
    remove_mono=False,
    title=None,
    rlabel=None,
    llabel=None,
    width=None,
    fraction=1,
    xsize=1000,
    darkmode=False,
    rot=None,
    coord=None,
    nest=False,
    flip="astro",
    graticule=False,
    graticule_labels=False,
    return_only_data=False,
    projection_type="mollweide",
    cb_orientation="horizontal",
    xlabel=None,
    ylabel=None,
    longitude_grid_spacing=60,
    latitude_grid_spacing=30,
    override_plot_properties=None,
    xtick_label_color="black",
    ytick_label_color="black",
    graticule_color=None,
    fontsize=None,
    phi_convention="counterclockwise",
    custom_xtick_labels=None,
    custom_ytick_labels=None,
    ratio=None,
    extend=None,
    return_figure=False,
    scale=1,
    **kwargs,
):
    """
    General plotting function for healpix maps.
    This function is a wrapper on healpys projview function with some added features.
    Features added in addition to existing projview features include:
        - direct plotting from model object, fits file or numpy array.
        - simple map operations such as smoothing, ud_grading, masking
        - removing dipole or monopoles
        - Predefined plotting parameters for specific components for nice formatting

    Parameters
    ----------
    input : ndarray, fits file path or cosmoglobe model object
        Map data input given as numpy array either 1d or index given by 'sig'.
        Also supports fits-file path string or cosmoglobe model.
        If cosmoglobe object is passed such as 'model', specify comp or freq.
    sig : str or int, optional
        Specify which signal to plot if ndim>1.
        default: None
    comp : string, optional
        Component label for automatic identification of plotting
        parameters based on information from autoparams.json
        default: None
    freq : astropy GHz, optional
        frequency in GHz needed for scaling maps when using a model object input
        default: None
    ticks : list or str, optional
        Min and max value for data. If None, uses 97.5th percentile.
        default: None
    min : float, optional
      The minimum range value. If specified, overwrites autodetector.
      default: None
    max : float, optional
      The maximum range value. If specified, overwrites autodetector.
      default: None
    rng : float, optional
      Sets this value as min and max value. If specified, overwrites autodetector.
      default: None
    cbar : bool, optional
        Toggles the colorbar
        cbar : True
    fwhm : astropy arcmin/rad/deg, optional
        Optional map smoothing. FWHM of gaussian smoothing in arcmin.
        default: 0.0
    mask : str path or np.ndarray, optional
        Apply a mask file to data
        default: None
    cmap : str, optional
        Colormap (ex. sunburst, planck, jet). Both matplotliib and cmasher
        available as of now. Also supports qualitative plotly map, [ex.
        q-Plotly-4 (q for qualitative 4 for max color)] Sets planck as default.
        default: None
    norm : {'hist', 'log', 'symlog', 'symlog2', None}
      Color normalization:
      hist = histogram equalized color mapping.
      log = logarithmic color mapping.
      symlog = symmetric logarithmic, linear between -linthresh and linthresh.
      symlog2 = similar to symlog, used for plack log colormap.
      default: None (linear color mapping)
    norm_dict : dict, optionals
        Parameters for normalization:
        default is set to {"linthresh": 1, "base": 10, "linscale": 0.1}
        where linthresh determines the linear regime of symlog norm,
        and linscale sets the size of the linear regime on the cbar.
        default: None
    remove_dip : bool, optional
        If mdmask is specified, fits and removes a dipole.
        default: True
    remove_mono : bool, optional
        If mdmask is specified, fits and removes a monopole.
        default: True
    unit : str, optional
        Unit label for colorbar
        default: None
    title : str, optional
        Sets the full figure title. Has LaTeX functionaliity (ex. $A_{s}$.)
        default: None
    rlabel : str, optional
        Sets the upper right title. Has LaTeX functionaliity (ex. $A_{s}$.)
        default: None
    llabel : str, optional
        Sets the upper left title. Has LaTeX functionaliity (ex. $A_{s}$.)
        default: None
    width : float, optional
        Size in inches default is half latex page, see fraction.
        default: None
    fraction : float, optional
        Fraction of latex page width.
        default : 1
    darkmode : bool, optional
        Plots all outlines in white for dark backgrounds, and adds 'dark' in
        filename.
        default: False
    rot : scalar or sequence, optional
      Describe the rotation to apply.
      In the form (lon, lat, psi) (unit: degrees) : the point at
      longitude *lon* and latitude *lat* will be at the center. An additional rotation
      of angle *psi* around this direction is applied.
    coord : sequence of character, optional
      Either one of 'G', 'E' or 'C' to describe the coordinate
      system of the map, or a sequence of 2 of these to rotate
      the map from the first to the second coordinate system.
    nest : bool, optional
      If True, ordering scheme is NESTED. Default: False (RING)
    flip : {'astro', 'geo'}, optional
      Defines the convention of projection : 'astro' (default, east towards left, west towards right)
      or 'geo' (east towards roght, west towards left)
      It creates the `healpy_flip` attribute on the Axes to save the convention in the figure.
    graticule : bool
      add graticule
    graticule_labels : bool
      longitude and latitude labels
    return_only_data : bool
      Return figure
    projection_type :  {'aitoff', 'hammer', 'lambert', 'mollweide', 'cart', '3d', 'polar'}
      type of the plot
    cb_orientation : {'horizontal', 'vertical'}
      color bar orientation
    xlabel : str
      set x axis label
    ylabel : str
      set y axis label
    longitude_grid_spacing : float
      set x axis grid spacing
    latitude_grid_spacing : float
      set y axis grid spacing
    override_plot_properties : dict
      Override the following plot proporties: 'cbar_shrink', 'cbar_pad', 'cbar_label_pad', 'figure_width': width, 'figure_size_ratio': ratio.
    lcolor : str
      change the color of the longitude tick labels, some color maps make it hard to read black tick labels
    fontsize:  dict
        Override fontsize of labels: 'xlabel', 'ylabel', 'title', 'xtick_label', 'ytick_label', 'cbar_label', 'cbar_tick_label'.
        default = None
    phi_convention : string
        convention on x-axis (phi), 'counterclockwise' (default), 'clockwise', 'symmetrical' (phi as it is truly given)
        if `flip` is 'geo', `phi_convention` should be set to 'clockwise'.
    custom_xtick_labels : list
        override x-axis tick labels
    custom_ytick_labels : list
        override y-axis tick labels
    scale : float
        rescales data by factor scale
    kwargs : keywords
        passed to projview
    """


    if not fontsize:
        fontsize = DEFAULT_FONTSIZES
    else:
        fontsize_ = DEFAULT_FONTSIZES.copy()
        for key in fontsize.keys():
            fontsize_[key] = fontsize[key]
        fontsize = fontsize_

    # Get figure width from page fraction
    if width is None:
        figsize = get_figure_width(fraction=fraction)
        width = figsize[0]

    set_style(darkmode)

    # Translate sig to correct format
    if isinstance(sig, str):
        sig = STOKES.index(sig)

    # Get data
    m, comp, freq, nside = get_data(input, sig, comp, freq, fwhm, nside=nside,
        sample=sample, scale=scale, remove_dip=remove_dip,
        remove_mono=remove_mono)

    if mask is not None:
        if isinstance(mask, str):
            mask = hp.read_map(mask)

        if hp.get_nside(mask) != nside:
            print("[magenta]Input mask nside is different, ud_grading to output nside.[/magenta]")
            mask = hp.ud_grade(mask, nside)
        m.mask = np.logical_not(mask)

    # Pass all your arguments in, return parsed plotting parameters
    params = get_params(
        data=m,
        comp=comp,
        sig=sig,
        rlabel=rlabel,
        llabel=llabel,
        unit=unit,
        ticks=ticks,
        min=min,
        max=max,
        rng=rng,
        norm=norm,
        norm_dict=norm_dict,
        cmap=cmap,
        freq_ref=freq,
        width=width,
        nside=nside,
    )

    # Colormap
    if not isinstance(cmap, colors.Colormap):
        cmap = load_cmap(params["cmap"])
    if maskfill:
        cmap.set_bad(maskfill)

    if override_plot_properties is None:
        override_plot_properties = {"cbar_tick_direction": "in"}

    warnings.filterwarnings("ignore")  # Healpy complains too much
    # Plot figure
    ret = temp_projview(
        params["data"],
        min=np.min(params["ticks"]),
        max=np.max(params["ticks"]),
        cbar_ticks=params["ticks"],
        cbar=cbar,
        cmap=cmap,
        unit=params["unit"],
        llabel=params["llabel"],
        rlabel=params["rlabel"],
        norm=params["norm"],
        norm_dict=params["norm_dict"],
        override_plot_properties=override_plot_properties,
        show_tickmarkers=True,
        width=width,
        # unedited params
        xsize=xsize,
        title=title,
        rot=rot,
        coord=coord,
        nest=nest,
        flip=flip,
        graticule=graticule,
        graticule_labels=graticule_labels,
        return_only_data=return_only_data,
        projection_type=projection_type,
        cb_orientation=cb_orientation,
        xlabel=xlabel,
        ylabel=ylabel,
        longitude_grid_spacing=longitude_grid_spacing,
        latitude_grid_spacing=latitude_grid_spacing,
        xtick_label_color=xtick_label_color,
        ytick_label_color=ytick_label_color,
        graticule_color=graticule_color,
        fontsize=fontsize,
        phi_convention=phi_convention,
        custom_xtick_labels=custom_xtick_labels,
        custom_ytick_labels=custom_ytick_labels,
        extend=extend,
        **kwargs
    )

    if return_figure:
       return ret, params



class ChainKeyError(Exception):
    """Raised when a key is missing from the chain."""


class ChainSampleError(Exception):
    """Raised when a sample is missing from the chain."""


class ChainComponentNotFoundError(Exception):
    """Raised when a component is missing from the chain."""


class ChainFormatError(Exception):
    """Raised when there is a format error with the chain."""



from matplotlib import rcParams
import matplotlib.patheffects as path_effects
import matplotlib.ticker as ticker
import matplotlib.pyplot as plt
from cycler import cycler

import healpy as hp
import matplotlib.pyplot as plt

from .plottools import *

from cosmoglobe.sky._units import Unit
import astropy.constants as c
import astropy.units as u


# TODO:
# CO is currently hardcoded
# Make better custom band selection

# Convert from RJ to MJy/sr; 2*nu**2*k_B/c**2 * I_nu


def spec(
    model,
    pol=False,
    sky_fractions=(25, 85),
    xlim=(0.25, 4000),
    figsize=None,
    fraction=1,
    labelsize=11,
    comp_labelsize=9,
    band_labelsize=8,
    extend=True,
    darkmode=False,
    ame_polfrac=0.006,
    haslam=True,
    chipass=True,
    spass=True,
    cbass=True,
    quijote=False,
    wmap=True,
    planck=True,
    dirbe=True,
    litebird=False,
    SO=False,
    S4=False,
    custom_bands=None,
    include_co=True,
    add_error=True,
    fsky_pos=500,
    component_options=None,
    text_offset=2,
    unit='uK_RJ',
):
    """
    Generates RMS intensity plot for components in model.
        `model = cosmoglobe.model_from_chain(chain, nside=64, components=["synch", "ff", "ame", "dust", "radio"])`
        `spec(model)`

    Parameters
    ----------
    model : cosmoglobe Model object
        Object containing all sky components to be visualized
    pol : bool, optional
        Plot polarization instead of intensity
        default = None (temperature)
    sky_fractions : touple, optional
        Min and max sky fractions for thickness of lines
        default: (25,85)
    xlim : touple, optional
        The limits of the x axis
        default: (0.25, 4000)
    figsize : touple, optional
        size of figure. This is automatically set by extend.
        default : None
    fraction : float, optional
        Sets the figsize to be a fraction of a latex document
        default : None
    labelsize : float,
        Fontsize of ticklabels and axislabels
        default : 11
    comp_labelsize : float,
        Fontsize of component labels
        default : 9
    band_labelsize : float,
        Fontsize of frequency band labels
        default : 8
    extend: bool, optional
        Increases the xlimit and includes an additional subplot on top
        default: True
    darkmode : bool, optional
        Set True to turn all frame elements white for black background viewing.
        default: False
    ame_polfrac: float, optional
        Set the polarization fraction of AME with a scale factor.
        default: 0.006
    haslam : bool, optional
        Toggle this observation column
        default:  True
    chipass : bool, optional
        Toggle this observation column
        default:  True
    spass : bool, optional
        Toggle this observation column
        default:  True
    cbass : bool, optional
        Toggle this observation column
        default:  True
    quijote : bool, optional
        Toggle this observation column
        default:  False
    wmap : bool, optional
        Toggle this observation column
        default:  True
    planck : bool, optional
        Toggle this observation column
        default:  True
    dirbe : bool, optional
        Toggle this observation column
        default:  True
    litebird : bool, optional
        Toggle this observation column
        default:  False
    SO : bool, optional
        Toggle this observation column
        default:  False
    S4 : bool, optional
        Toggle this observation column
        default:  False
    custom_bands : bool, optional
        TODO: NOT IMPLEMENTED, toggle specific bands
        default:  None
    include_co : bool,  optional
        Turn off CO lines
        default: True
    add_error : bool, optional
        Toggle added uncertainty to low S/N part of spectrum.
        default: True
    fsky_pos : float, optional
        Frequency position of f_sky labels
        default : 900
    component_options : dict, optional
        Allows for overriding dictionary that sets all the properties of components.
        Such as {"dust":{"position":30, "label": "Thing", "linestyle": ":"}}
        default : None
    text_offset : float, optional
        Text offset from lines
        default : 5
    unit: uK_RJ or MJy/sr
    """
    if fraction is not None:
        figsize = get_figure_width(fraction=fraction)
    set_style(darkmode, font="serif")

    # rcParams.update(figparams)
    black = "k"
    if darkmode:
        brokecol = "white"
        grey = "#C0C0C0"
    else:
        brokecol = black
        grey = "#A3A3A3"

    colors = [
        "#636EFA",
        "#EF553B",
        "#00CC96",
        "#AB63FA",
        "#FFA15A",
        "#19D3F3",
        "#FF6692",
        "#B6E880",
        "#FF97FF",
        "#FECB52",
        grey,
    ]
    rcParams["axes.prop_cycle"] = cycler(color=colors)
    rcParams["ytick.right"] = True

    blue, red, green, purple, orange, teal, lightred, lightgreen, pink, yellow, grey = (
        "C0",
        "C1",
        "C2",
        "C3",
        "C4",
        "C5",
        "C6",
        "C7",
        "C8",
        "C9",
        "C10",
    )

    # plt.rcParams.update(plt.rcParamsDefault)

    # Plot polarization or temperature
    sig = 1 if pol else 0

    # Set x and y limits if extended and pol
    if xlim == (0.25, 4000):
        if not extend:
            xlim = (9, 1500)
    xmin, xmax = xlim
    ymin, ymax = (0.05, 7e2) if not pol else (1.001e-3, 2e2)
    ymin2, ymax2 = (ymax + 100, 1e7)

    figsize = get_figure_width(fraction=fraction) if figsize is None else figsize

    if extend:
        ratio = 5  # Make subplot on top at 5:1 ratio
        fig, (ax2, ax) = plt.subplots(
            2,
            1,
            sharex=True,
            figsize=(figsize[0], figsize[1]),
            gridspec_kw={"height_ratios": [1, ratio]},
        )
        aspect_ratio = figsize[0] / figsize[1] * 1.25  # Correct for ratio

        ax2.spines["bottom"].set_visible(False)
        ax.spines["top"].set_visible(False)
        ax2.tick_params(labelbottom=False)
        ax2.xaxis.set_ticks_position("none")

        # ---- Adding broken axis lines ----
        d = 0.005  # how big to make the diagonal lines in axes coordinates

        kwargs = dict(transform=ax2.transAxes, color=brokecol, clip_on=False, linewidth=1)
        ax2.plot((-d, +d), (-d * ratio, +d * ratio), **kwargs)  # top-left diagonal
        ax2.plot((1 - d, 1 + d), (-d * ratio, +d * ratio), **kwargs)  # top-right diagonal
        kwargs.update(transform=ax.transAxes)  # switch to the bottom axes
        ax.plot((-d, +d), (1 - d, 1 + d), **kwargs)  # bottom-left diagonal
        ax.plot((1 - d, 1 + d), (1 - d, 1 + d), **kwargs)  # bottom-right diagonal

    else:
        ymax2 = ymax
        ymin2 = ymax
        fig, ax = plt.subplots(1, 1, figsize=figsize)
        aspect_ratio = figsize[0] / figsize[1]

    # Spectrum parameters
    N = 1000
    nu = np.logspace(np.log10(0.1), np.log10(300_000), N)
    
    if unit == 'uK_RJ':
        C_fact = 1
    else:
        C_fact = (2*(nu*u.GHz)**2*c.k_B/c.c**2).to('MJy uK-1').value

    seds = seds_from_model(nu, model, pol=pol, sky_fractions=sky_fractions)

    # Get parameters for each component line
    foregrounds = get_foregrounds(pol, extend)
    if component_options is not None:
        # Loop over nested dictionary for optional arguments
        for comp_ in component_options:
            foregrounds[comp_].update(component_options[comp_])

    # Looping over foregrounds and calculating spectra
    i = 0
    for comp in foregrounds.keys():
        if comp.startswith("co") and include_co:  # get closest thing to ref freq
            foregrounds[comp]["params"][2], line_idx = find_nearest(nu, foregrounds[comp]["params"][2])
            foregrounds[comp]["spectrum"] = np.zeros((2, len(sky_fractions), N))
            foregrounds[comp]["spectrum"][sig][0][line_idx] = foregrounds[comp]["params"][0]
            foregrounds[comp]["spectrum"][sig][1][line_idx] = foregrounds[comp]["params"][1]

        if comp.startswith("bb"):
            a = 0.67 * 1e-1 if comp.endswith("2") else 0.67 * 1e-2
            sed = np.zeros((2, len(sky_fractions), N))
            cmb_blackbody = (np.ones(len(nu)) * Unit("uK_CMB")).to("uK_RJ", equivalencies=cmb_equivalencies(nu * u.GHz))
            sed[1] = a * cmb_blackbody
            foregrounds[comp]["spectrum"] = sed

        if comp in seds.keys():
            foregrounds[comp]["spectrum"] = seds[comp]
        else:
            continue

        if pol and comp == "ame":
            foregrounds[comp]["spectrum"][1] = ame_polfrac * foregrounds[comp]["spectrum"][0]

        if add_error and not comp.startswith("co") and not comp.startswith("bb") and not comp.startswith("cmb"):
            thresh = 0.1
            alpha = 0.5
            foregrounds[comp]["spectrum"][sig][0] = foregrounds[comp]["spectrum"][sig][0] * (1 - np.exp(-(abs(foregrounds[comp]["spectrum"][sig][0] / thresh) ** alpha)))
            foregrounds[comp]["spectrum"][sig][1] = foregrounds[comp]["spectrum"][sig][1] / (1 - np.exp(-(abs(foregrounds[comp]["spectrum"][sig][1] / thresh) ** alpha)))
            foregrounds[comp]["spectrum"][sig][0][np.isnan(foregrounds[comp]["spectrum"][sig][0])] = 0
            foregrounds[comp]["spectrum"][sig][1][np.isnan(foregrounds[comp]["spectrum"][sig][1])] = 0

        if foregrounds[comp]["sum"] and foregrounds[comp]["spectrum"] is not None:
            if i == 0:
                foregrounds["sumfg"]["spectrum"] = foregrounds[comp]["spectrum"].copy()
            else:
                foregrounds["sumfg"]["spectrum"] += foregrounds[comp]["spectrum"]
            i += 1

    # ---- Plotting foregrounds and labels ----
    j = 0
    for comp, params in foregrounds.items():  # Plot all fgs except sumf
        if params["spectrum"] is None and not comp.startswith("co"):
            continue
        if not params["gradient"]:
            k = 1
            gradient_fill_between(
                ax,
                nu,
                params["spectrum"][sig][1] * 1e-2,
                params["spectrum"][sig][1],
                color=params["color"],
            )
        else:
            if comp == "sumfg":
                ax.loglog(
                    nu,
                    params["spectrum"][sig][1]*C_fact,
                    linestyle=params["linestyle"],
                    linewidth=1.5,
                    color=params["color"],
                )
                if extend:
                    ax2.loglog(
                        nu,
                        params["spectrum"][sig][1]*C_fact,
                        linestyle=params["linestyle"],
                        linewidth=1.5,
                        color=params["color"],
                    )
                k = 1
                try:
                    ax.loglog(
                        nu,
                        params["spectrum"][sig][0]*C_fact,
                        linestyle=params["linestyle"],
                        linewidth=1.5,
                        color=params["color"],
                    )
                    if extend:
                        ax2.loglog(
                            nu,
                            params["spectrum"][sig][0]*C_fact,
                            linestyle=params["linestyle"],
                            linewidth=1.5,
                            color=params["color"],
                        )
                    k = 1
                except:
                    pass
            elif comp.startswith("co"):
                if include_co:
                    ax.loglog(
                        [params["params"][2], params["params"][2]],
                        [
                            max(params["spectrum"][sig][0]),
                            max(params["spectrum"][sig][1]),
                        ],
                        linestyle=params["linestyle"],
                        linewidth=2,
                        color=params["color"],
                        zorder=1000,
                    )
                    k = 1
                else:
                    continue
            else:
                if comp == "cmb":
                    ax.loglog(
                        nu,
                        params["spectrum"][sig][0]*C_fact,
                        linestyle=params["linestyle"],
                        linewidth=2,
                        color=params["color"],
                    )
                    if extend:
                        ax2.loglog(
                            nu,
                            params["spectrum"][sig][0]*C_fact,
                            linestyle=params["linestyle"],
                            linewidth=4,
                            color=params["color"],
                        )
                    k = 0
                else:
                    ax.fill_between(
                        nu,
                        params["spectrum"][sig][1]*C_fact,
                        params["spectrum"][sig][0]*C_fact,
                        color=params["color"],
                        alpha=0.8,
                    )
                    if extend:
                        ax2.fill_between(
                            nu,
                            params["spectrum"][sig][1]*C_fact,
                            params["spectrum"][sig][0]*C_fact,
                            color=params["color"],
                            alpha=0.8,
                        )
                    k = 1

        if comp.startswith("co") and include_co:
            ax.text(
                foregrounds[comp]["params"][2],
                np.max(params["spectrum"][sig][k]) * 0.5,
                params["label"],
                color=params["color"],
                alpha=0.7,
                ha="right",
                va="center",
                rotation=90,
                fontsize=comp_labelsize,
                path_effects=[path_effects.withSimplePatchShadow(alpha=0.8, offset=(0.3, -0.3))],
                zorder=1000,
            )
        else:
            # Label each component with rotation
            x0, y0, rotation = rotate_label(
                nu,
                params["spectrum"][sig][k],
                params["position"],
                (xmin, xmax),
                (ymin, ymax),
                aspect_ratio,
            )
            _text_offset = text_offset * 1.5 if comp == "sumfg" else text_offset
            ax.annotate(
                params["label"],
                xy=(x0, y0),
                xytext=(0, _text_offset),
                textcoords="offset pixels",
                rotation=rotation,
                rotation_mode="anchor",
                fontsize=comp_labelsize,
                color=params["color"],
                path_effects=[
                    path_effects.withSimplePatchShadow(alpha=0.8, offset=(0.3, -0.3)),
                ],
                horizontalalignment="center",
            )

            if comp == "dust":
                # Label the sky-fraction around the dust line
                x0, y0, rotation = rotate_label(
                    nu,
                    params["spectrum"][sig][1],
                    fsky_pos,
                    (xmin, xmax),
                    (ymin, ymax),
                    aspect_ratio,
                )
                ax.annotate(
                    r"$f_{\mathrm{sky}}=" + "{:d}".format(int(sky_fractions[1])) + "\%$",
                    rotation=rotation,
                    rotation_mode="anchor",
                    xy=(x0, y0),
                    ha="center",
                    va="bottom",
                    fontsize=comp_labelsize,
                    color=grey,
                    xytext=(0, text_offset),
                    textcoords="offset pixels",
                    path_effects=[
                        path_effects.withSimplePatchShadow(alpha=0.8, offset=(0.3, -0.3)),
                    ],
                )

                x0, y0, rotation = rotate_label(
                    nu,
                    params["spectrum"][sig][0],
                    fsky_pos,
                    (xmin, xmax),
                    (ymin, ymax),
                    aspect_ratio,
                )
                ax.annotate(
                    r"$f_{\mathrm{sky}}=" + "{:d}".format(int(sky_fractions[0])) + "\%$",
                    rotation=rotation,
                    rotation_mode="anchor",
                    xy=(x0, y0),
                    ha="center",
                    va="top",
                    fontsize=comp_labelsize,
                    color=grey,
                    xytext=(0, -text_offset),
                    textcoords="offset pixels",
                    path_effects=[
                        path_effects.withSimplePatchShadow(alpha=0.8, offset=(0.3, -0.3)),
                    ],
                )

    # ---- Data band ranges ----
    if extend:
        yscaletext = 0.70
        yscaletextup = 1.2
    else:
        yscaletextup = 1.03
        yscaletext = 0.90

    if custom_bands is not None:
        """
        TODO: Make it easier to specify which bands to include
        """
        pass
    if isinstance(planck, bool):
        planck = [planck] * 9
    if isinstance(wmap, bool):
        wmap = [wmap] * 5

    # TODO add these as args?
    # fmt: off
    databands = {
        "Haslam": {
            "0.408\nHaslam": { "pol": False, "show": haslam, "position": [0.408, ymin * yscaletextup], "range": [0.406, 0.410], "color": purple, }
        },
        "S-PASS": {
            "2.303\nS-PASS": { "pol": True, "show": spass, "position": [2.35, ymax2 * yscaletext], "range": [2.1, 2.4], "color": green, }
        },
        "C-BASS": {
            "5.0\nC-BASS": { "pol": True, "show": cbass, "position": [5.0, ymax2 * yscaletext], "range": [4.0, 6.0], "color": blue, }
        },
        "CHI-PASS": {
            "1.394\nCHI-PASS": { "pol": False, "show": chipass, "position": [1.3945, ymin * yscaletextup], "range": [1.3945 - 0.064 / 2, 1.3945 + 0.064 / 2], "color": lightred, }
        },
        "QUIJOTE": {
            "11\nQUIJOTE": { "pol": True, "show": quijote, "position": [11, ymax2 * yscaletext], "range": [10.0, 12.0], "color": red,},
            "13": { "pol": True, "show": quijote, "position": [13, ymax2 * yscaletext], "range": [12.0, 14.0], "color": red,},
            "17": { "pol": True, "show": quijote, "position": [17, ymax2 * yscaletext], "range": [16.0, 18.0], "color": red,},
            "19": { "pol": True, "show": quijote, "position": [20, ymax2 * yscaletext], "range": [18.0, 21.0], "color": red,},
            ".\n31": { "pol": True, "show": quijote, "position": [31, ymax2 * yscaletext], "range": [26.0, 36.0], "color": red,},
            ".\n41": { "pol": True, "show": quijote, "position": [42, ymax2 * yscaletext], "range": [35.0, 47.0], "color": red,},
        },
        "Planck": {
            "30": { "pol": True, "show": planck[0], "position": [28, ymax2 * yscaletext], "range": [23.9, 34.5], "color": orange,},  # Planck 30
            "44": { "pol": True, "show": planck[1], "position": [44, ymax2 * yscaletext], "range": [39, 50], "color": orange,},  # Planck 44
            "70": { "pol": True, "show": planck[2], "position": [70, ymax2 * yscaletext], "range": [60, 78], "color": orange,},  # Planck 70
            "100": { "pol": True, "show": planck[3], "position": [100, ymax2 * yscaletext], "range": [82, 120], "color": orange,},  # Planck 100
            "143": { "pol": True, "show": planck[4], "position": [145, ymax2 * yscaletext], "range": [125, 170], "color": orange,},  # Planck 143
            "217": { "pol": True, "show": planck[5], "position": [217, ymax2 * yscaletext], "range": [180, 265], "color": orange,},  # Planck 217
            "353": { "pol": True, "show": planck[6], "position": [350, ymax2 * yscaletext], "range": [300, 430], "color": orange,},  # Planck 353
            "545": { "pol": False, "show": planck[7], "position": [540, ymax2 * yscaletext], "range": [450, 650], "color": orange,},  # Planck 545
            "857": { "pol": False, "show": planck[8], "position": [850, ymax2 * yscaletext], "range": [700, 1020], "color": orange,},
        },  # Planck 857
        "DIRBE": {
            "DIRBE\n1250": { "pol": False, "show": dirbe, "position": [1200, ymin * yscaletextup], "range": [1000, 1540], "color": red,},  # DIRBE 1250
            "2140": { "pol": False, "show": dirbe, "position": [1900, ymin * yscaletextup], "range": [1780, 2500], "color": red,},  # DIRBE 2140
            "3000": { "pol": False, "show": dirbe, "position": [3000, ymin * yscaletextup], "range": [2600, 3500], "color": red,},
        },  # DIRBE 3000
        "WMAP": {
            "K": { "pol": True, "show": wmap[0], "position": [23, ymin * yscaletextup], "range": [21, 25.5], "color": teal,},
            "Ka": { "pol": True, "show": wmap[1], "position": [33, ymin * yscaletextup], "range": [30, 37], "color": teal,},
            "Q": { "pol": True, "show": wmap[2], "position": [41.0, ymin * yscaletextup], "range": [38, 45], "color": teal,},
            "V": { "pol": True, "show": wmap[3], "position": [60.0, ymin * yscaletextup], "range": [54, 68], "color": teal,},
            "W": { "pol": True, "show": wmap[4], "position": [90.0, ymin * yscaletextup], "range": [84, 106], "color": teal,},
        },
        "LiteBIRD": {
            "LiteBIRD 40": { "pol": True, "show": litebird, "position": [33, ymax2 * yscaletext], "range": [34, 46], "color": red,},
            "50": { "pol": True, "show": litebird, "position": [50, ymax2 * yscaletext], "range": [43, 57], "color": red,},
            "60": { "pol": True, "show": litebird, "position": [60, ymax2 * yscaletext], "range": [53, 67], "color": red,},
            "68": { "pol": True, "show": litebird, "position": [68, ymax2 * yscaletext], "range": [60, 76], "color": red,},
            "78": { "pol": True, "show": litebird, "position": [78, ymax2 * yscaletext], "range": [69, 87], "color": red,},
            "89": { "pol": True, "show": litebird, "position": [89, ymax2 * yscaletext], "range": [79, 99], "color": red,},
            "100": { "pol": True, "show": litebird, "position": [100, ymax2 * yscaletext], "range": [89, 111], "color": red,},
            "119": { "pol": True, "show": litebird, "position": [119, ymax2 * yscaletext], "range": [101, 137], "color": red,},
            "140": { "pol": True, "show": litebird, "position": [140, ymax2 * yscaletext], "range": [119, 161], "color": red,},
            "166": { "pol": True, "show": litebird, "position": [166, ymax2 * yscaletext], "range": [141, 191], "color": red,},
            "195": { "pol": True, "show": litebird, "position": [195, ymax2 * yscaletext], "range": [165, 225], "color": red,},
            "235": { "pol": True, "show": litebird, "position": [235, ymax2 * yscaletext], "range": [200, 270], "color": red,},
            "280": { "pol": True, "show": litebird, "position": [280, ymax2 * yscaletext], "range": [238, 322], "color": red,},
            "337": { "pol": True, "show": litebird, "position": [337, ymax2 * yscaletext], "range": [287, 387], "color": red,},
            "402": { "pol": True, "show": litebird, "position": [402, ymax2 * yscaletext], "range": [356, 458], "color": red,},
        },
        "SO" : {
            "SO 27": { "pol": True, "show": SO, "position":[25, ymax2*yscaletext], "range": [23, 30], "color":green,},
            "39": { "pol": True, "show": SO, "position":[39, ymax2*yscaletext], "range": [32, 49], "color":green,},
            "93": { "pol": True, "show": SO, "position":[93, ymax2*yscaletext], "range": [75, 120], "color":green,},
            "145": { "pol": True, "show": SO, "position":[145, ymax2*yscaletext], "range": [130, 180], "color":green,},
            "225": { "pol": True, "show": SO, "position":[225, ymax2*yscaletext], "range": [190, 260], "color":green,},
            "280": { "pol": True, "show": SO, "position":[280, ymax2*yscaletext], "range": [260, 320], "color":green,},
        },
        "S4-SAT" : { 
            "S4-SAT 30": { "pol": True, "show": S4, "position":[24, ymax2*yscaletext], "range": [26, 35], "color":yellow,},
            "40": { "pol": True, "show": S4, "position":[40, ymax2*yscaletext], "range": [34, 46], "color":yellow,},
            "85": { "pol": True, "show": S4, "position":[85, ymax2*yscaletext], "range": [75, 95], "color":yellow,}, 
            "95": { "pol": True, "show": S4, "position":[95, ymax2*yscaletext], "range": [84, 106], "color":yellow,},
            "145": { "pol": True, "show": S4, "position":[130, ymax2*yscaletext], "range": [129, 161], "color":yellow,},
            "155": { "pol": True, "show": S4, "position":[155, ymax2*yscaletext], "range": [138, 172], "color":yellow,},
            "220": { "pol": True, "show": S4, "position":[220, ymax2*yscaletext], "range": [196, 244], "color":yellow,},
            "270": { "pol": True, "show": S4, "position":[270, ymax2*yscaletext], "range": [240, 300], "color":yellow,},

        },
        "S4-LAT" : { 
            "20": { "pol": True, "show": S4, "position":[20, ymax2*yscaletext], "range": [15, 25], "color":purple,},
            "27": { "pol": True, "show": S4, "position":[27, ymax2*yscaletext], "range": [24, 30], "color":purple,},
            "39": { "pol": True, "show": S4, "position":[39, ymax2*yscaletext], "range": [21, 57], "color":purple,},
            "93\nS4-LAT": { "pol": True, "show": S4, "position":[93, ymax2*yscaletext], "range": [75, 110], "color":purple,},
            "145": { "pol": True, "show": S4, "position":[145, ymax2*yscaletext], "range": [125, 165], "color":purple,},
            "225" : { "pol": True, "show": S4, "position":[225, ymax2*yscaletext], "range": [195, 255], "color":purple,},
            "278" : { "pol": True, "show": S4, "position":[278, ymax2*yscaletext], "range": [256, 300], "color":purple,},
        },
    }
    # fmt: on
    
    # If band is not included, label another band:
    if planck[4]:
        oldkey = list(databands["Planck"].keys())[3]
    else:
        for i in [3, 2, 1, 5, 6, 7, 8]:
            if planck[i]:
                oldkey = list(databands["Planck"].keys())[i]
                break

    if(litebird):
        oldkey = list(databands["Planck"].keys())[8]

    newkey = oldkey + "\nPlanck"
    databands["Planck"][newkey] = databands["Planck"].pop(oldkey)

    # Same for wmap
    if wmap[1]:
        oldkey = list(databands["WMAP"].keys())[1]
    else:
        for i in [1, 2, 0, 3, 4]:
            if planck[i]:
                oldkey = list(databands["Planck"].keys())[i]
                break

    newkey = "WMAP\n" + oldkey
    databands["WMAP"][newkey] = databands["WMAP"].pop(oldkey)

    # Set databands from dictonary
    v_pos = ymax2
    for experiment, bands in databands.items():
        if experiment in ["LiteBIRD", "SO", "S4-LAT", "S4-SAT"]:
            v_pos = v_pos * 0.2  #add vertical space at the top to avoid the other labels
        for label, band in bands.items():
            if band["show"]:
                # if label == "353" and not pol: continue # SPECIFIC FOR BP SETUP
                if pol and not band["pol"]:
                    continue  # Skip non-polarization bands
                if band["position"][0] >= xmax or band["position"][0] <= xmin:
                    continue  # Skip databands outside range
                va = "bottom" if experiment in ["WMAP", "CHI-PASS", "DIRBE", "Haslam"] else "top"  # VA for WMAP on bottom
                if experiment in ["Planck", "LiteBIRD", "SO", "S4-LAT", "S4-SAT"]:
                    band["position"][1] = v_pos

                ha = (
                    "center"
                    if experiment
                    in [
                        "Planck",
                        "WMAP",
                        "DIRBE",
                    ]
                    else "center"
                )
                ax.axvspan(*band["range"], color=band["color"], alpha=0.3, zorder=-20, label=experiment)
                if extend:
                    ax2.axvspan(*band["range"], color=band["color"], alpha=0.3, zorder=-20, label=experiment)
                    if experiment in ["WMAP", "CHI-PASS", "DIRBE", "Haslam"]:
                        ax.text(*band["position"], label, color=band["color"], va=va, ha=ha, size=band_labelsize, path_effects=[path_effects.withSimplePatchShadow(alpha=0.8, offset=(0.3, -0.3))])
                    else:
                        ax2.text(*band["position"], label, color=band["color"], va=va, ha=ha, size=band_labelsize, path_effects=[path_effects.withSimplePatchShadow(alpha=0.8, offset=(0.3, -0.3))])
                else:
                    ax.text(*band["position"], label, color=band["color"], va=va, ha=ha, size=band_labelsize, path_effects=[path_effects.withSimplePatchShadow(alpha=0.8, offset=(0.3, -0.3))])

    # ---- Axis stuff ----
    xticks = []
    xticks_ = [0.3, 1, 3, 10, 30, 100, 300, 1000, 3000]
    for i, xtick in enumerate(xticks_):
        if xtick >= xmin and xtick <= xmax:
            xticks.append(xtick)

    yticks = []
    yticks_ = [0.001, 0.01, 0.1, 1, 10, 100, 1000]
    for i, ytick in enumerate(yticks_):
        if ytick >= ymin and ytick <= ymax:
            yticks.append(ytick)

    if unit == 'uK_RJ':
        ax.set(
            xscale="log",
            yscale="log",
            ylim=(ymin, ymax),
            xlim=(xmin, xmax),
            yticks=yticks,
            yticklabels=yticks,
            xticks=xticks,
            xticklabels=xticks,
        )
    formatter = ticker.FuncFormatter(fmt)
    ax.xaxis.set_major_formatter(formatter)
    ax.yaxis.set_major_formatter(formatter)
    ax.tick_params(axis="both", which="both", labelsize=labelsize, direction="in")
    ax.tick_params(
        axis="y",
        labelrotation=90,
        labelsize=labelsize,
    )

    if extend:
        ax2.set(
            xscale="log",
            yscale="log",
            ylim=(ymin2, ymax2),
            xlim=(xmin, xmax),
            xticks=xticks,
            xticklabels=xticks,
        )
        ax2.yaxis.set_major_formatter(formatter)
        ax2.tick_params(axis="both", which="both", labelsize=labelsize, direction="in")
        ax2.tick_params(
            axis="y",
            labelrotation=90,
            labelsize=labelsize,
        )
        ax2.set_yticklabels(
            [fmt(x, 1) for x in ax2.get_yticks()],
            fontsize=labelsize,
            va="center",
        )

    # Why do i have to do this?
    ax.set_yticklabels(
        [fmt(x, 1) for x in ax.get_yticks()],
        fontsize=labelsize,
        va="center",
    )
    ax.set_xticklabels([fmt(x, 1) for x in ax.get_xticks()], fontsize=labelsize)

    # Axis labels
    sax = fig.add_subplot(111, frameon=False)
    sax.tick_params(
        labelcolor="none",
        top=False,
        bottom=False,
        left=True,
        right=False,
        width=0.0,
    )

    sax.set_ylabel(r"$\mathrm{RMS\ amplitude\ [}\mu\mathrm{K}_{\mathrm{RJ}}\mathrm{]}$", fontsize=labelsize, labelpad=0)
    sax.set_xlabel(r"$\mathrm{Frequency\ [GHz]}$", fontsize=labelsize)
    plt.subplots_adjust(wspace=0.0, hspace=0.02)


def rotate_label(x, y, pos, xrange, yrange, aspect_ratio):
    """
    This function takes a line and calculates position and rotation
    for a label on top of it.
    Takes into account logspace etc.
    """
    x0, idx1 = find_nearest(x, pos)
    idx2 = idx1 + 2
    x1 = x[idx2]
    y0 = y[idx1]
    y1 = y[idx2]
    datascaling = np.log(xrange[0] / xrange[1]) / np.log(yrange[0] / yrange[1])
    rotator = datascaling / aspect_ratio
    alpha = np.arctan(np.log(y1 / y0) / np.log(x1 / x0) * rotator)
    rotation = np.rad2deg(alpha)  # *rotator
    return x0, y0, rotation


def get_foregrounds(pol, extend):
    """
    This function grabs a dictionary of the default component values
    used for the spectrum plotting style.
    Override this by passing the component_option dictionary to spec
    """
    if pol:
        p = 0.6 if extend else 15
        sd = 2 if extend else 70
        return {
            "synch": {
                "label": "Synchrotron",
                "params": [],
                "position": 20,
                "color": "C2",
                "sum": True,
                "linestyle": "solid",
                "gradient": False,
                "spectrum": None,
            },
            "dust": {
                "label": "Thermal Dust",
                "params": [],
                "position": 250,
                "color": "C1",
                "sum": True,
                "linestyle": "solid",
                "gradient": False,
                "spectrum": None,
            },
            "sumfg": {
                "label": "Sum fg.",
                "params": [],
                "position": 70,
                "color": "C10",
                "sum": False,
                "linestyle": "--",
                "gradient": False,
                "spectrum": None,
            },
            "bb-2": {
                "label": r"BB $r=10^{-2}$",
                "params": [
                    0.01,
                    "BB",
                ],
                "position": p,
                "color": "C10",
                "sum": False,
                "linestyle": "dotted",
                "gradient": True,
                "spectrum": None,
            },
            "bb-4": {
                "label": r"BB $r=10^{-4}$",
                "params": [
                    1e-4,
                    "BB",
                ],
                "position": p,
                "color": "C10",
                "sum": False,
                "linestyle": "dotted",
                "gradient": True,
                "spectrum": None,
            },
            "cmb": {
                "label": "CMB EE",
                "params": [1, "EE"],
                "position": p,
                "color": "C5",
                "sum": False,
                "linestyle": "solid",
                "gradient": False,
                "spectrum": None,
            },
            "ame": {
                "label": "Spinning Dust",
                "params": [],
                "position": sd,
                "color": "C4",
                "sum": True,
                "linestyle": "solid",
                "gradient": True,
                "spectrum": None,
            },
        }
    else:
        p = 3 if extend else 65
        td = 10 if extend else 17
        return {
            "dust": {
                "label": "Thermal Dust",
                "params": [],
                "position": td,
                "color": "C1",
                "sum": True,
                "linestyle": "solid",
                "gradient": False,
                "spectrum": None,
            },
            "ff": {
                "label": "Free-Free",
                "params": [],
                "position": 50,
                "color": "C0",
                "sum": True,
                "linestyle": "solid",
                "gradient": False,
                "spectrum": None,
            },
            "ame": {
                "label": "Spinning Dust",
                "params": [],
                "position": p,
                "color": "C4",
                "sum": True,
                "linestyle": "solid",
                "gradient": False,
                "spectrum": None,
            },
            "synch": {
                "label": "Synchrotron",
                "params": [],
                "position": 170,
                "color": "C2",
                "sum": True,
                "linestyle": "solid",
                "gradient": False,
                "spectrum": None,
            },
            "co10": {
                "label": r"CO$_{1\rightarrow 0}$",
                "params": [0.5, 8, 115, 11.06],
                "position": p,
                "color": "C9",
                "sum": True,
                "linestyle": "solid",
                "gradient": False,
                "spectrum": None,
            },
            "co21": {
                "label": r"CO$_{2\rightarrow 1}$",
                "params": [0.3, 5, 230.0, 14.01],
                "position": p,
                "color": "C9",
                "sum": True,
                "linestyle": "solid",
                "gradient": False,
                "spectrum": None,
            },
            "co32": {
                "label": r"CO$_{3\rightarrow 2}$",
                "params": [0.3, 1, 345.0, 12.24],
                "position": p,
                "color": "C9",
                "sum": True,
                "linestyle": "solid",
                "gradient": False,
                "spectrum": None,
            },
            "sumfg": {
                "label": "Sum fg.",
                "params": [],
                "position": 25,
                "color": "C10",
                "sum": False,
                "linestyle": "--",
                "gradient": False,
                "spectrum": None,
            },
            "cmb": {
                "label": "CMB",
                "params": [1.0, "TT"],
                "position": 70,
                "color": "C5",
                "sum": False,
                "linestyle": "solid",
                "gradient": False,
                "spectrum": None,
            },
        }



from .skymap import plot
from .gnomonic import gnom
from .histogram import hist
from .traceplot import trace
from .spectrum import spec
from .plottools import standalone_colorbar
from .temp_newvisufunc import projview



#================================================================================
#
# Copyright (C) 2020 Institute of Theoretical Astrophysics, University of Oslo.
#
# This file is part of Commander3.
#
# Commander3 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Commander3 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Commander3. If not, see <https://www.gnu.org/licenses/>.
#
#================================================================================

import numpy as np


# k = 0 means autoselect k
def encode(data, k=0):

    assert(np.issubdtype(data.dtype, np.integer))
    #bitstream = bitarray()
    bitstream = ''

    offset = int(np.mean(data))
    data = data - offset
    data = np.insert(data, 0, offset)

    if(k < 1): #autoselect k
        k = 1
        mean = np.mean(np.abs(data[1:]))
        if mean > 0:
            k = int(np.log2(mean))

    #print(offset, data, len(data), np.mean(np.abs(data)), pow(2, k), k)

    m = pow(2,k)        

    for entry in data:

        negBit = '1'
        if entry > 0:
            negBit = '0'

        entry = abs(entry)

        q = int(entry/m)
        for i in range(q):    
            #bitstream.append(1)
            bitstream += '1'
        #bitstream.append(0)
        bitstream += '0' + negBit
        #bitstream.append(negBit)   
 
        binary = bin(entry)[2:].zfill(k)
        bits = binary[-k:]
        #for bit in bits:
        #    bitstream.append(int(bit))
        bitstream += bits

        #print(entry, bitstream, q, binary, bits, negBit)

    padding = 8 - len(bitstream) % 8
    bitstream += padding*'0'
    

    b = bytearray()
    for i in range(0, len(bitstream), 8):
        byte = bitstream[i:i+8]
        b.append(int(byte, 2))

    #return bitstream, k
    return b, k

def decode(data, k):
    data = bytearray(data)
    binary_txt = ''.join(bin(i)[2:].rjust(8,'0') for i in data)

    k = int(k[:])
    m = pow(2, k)
    outData = []
    i = 0

    while i<len(binary_txt):

        if(len(binary_txt) - i < 8 and '1' not in binary_txt[i:]):
            break

        num = 0
        negBit = 1

        while(binary_txt[i] == '1'):
            num += m
            i += 1
        i += 1 #skip the placeholder 0

        if(binary_txt[i] == '1'): #read the negative bit
            negBit = -1

        i += 1
  
        num += int(binary_txt[i:i+k], 2)
        i += k

        outData.append(num*negBit)

    return outData



import click
from rich import print
from astropy.io import fits

import healpy as hp
import numpy as np


@click.group()
def commands_fits():
    pass

@commands_fits.command()
@click.argument("input", type=click.STRING)
def printheader(input,):
    """
    Prints the header of a fits file.
    """
    with fits.open(input) as hdulist:
        hdulist.info()
        for hdu in hdulist:
            print(repr(hdu.header))

@commands_fits.command()
@click.argument("input", type=click.STRING)
def printdata(input,):
    """
    Prints the data of a fits file
    """
    with fits.open(input) as hdulist:
        hdulist.info()
        for hdu in hdulist:
            print(repr(hdu.data))


@commands_fits.command()
@click.argument("input", type=click.STRING)
@click.argument("output", type=click.STRING)
@click.argument("columns", type=click.INT)
def rmcolumn(input, output, columns):
    """
    Removes columns in fits file
    """
    with fits.open(input) as hdulist:
        for hdu in hdulist:
            hdu.header.pop(columns)        
            hdu.data.del_col(columns)        
        hdulist.writeto(output, overwrite=True)
        
# fmt: off
@commands_fits.command()
@click.argument("input", type=click.STRING)
@click.option("-mask", type=click.STRING, help="Mask for dipole removal, 30deg sky cut by default")
@click.option("-gal_cut", default=30, type=click.FLOAT, help="Latitude cut if no mask is provided")
@click.option("-sig", type=click.INT, multiple=True, help="fields to calculate, [0,1,2 by default]")
# fmt: on
def rmmd(input, mask, sig, gal_cut):
    """
    removes the dipole and mask of input fits file
    If mask is not specified, uses 30
    """

    try:
        m = hp.read_map(input, field=(0,1,2), verbose=False, dtype=None,)
    except:
        m = hp.read_map(input, field=(0,), verbose=False, dtype=None,)
    nside = hp.get_nside(m)
    npix = hp.nside2npix(nside)
    print(f"Nside: {nside}, npix: {npix}")

    # Mask map for dipole estimation
    m_masked = hp.ma(m)
    if mask:
        m_masked.mask = np.logical_not(hp.read_map(mask,verbose=False,dtype=None,))
    
    # Fit dipole to masked map
    for i in sig:
        if not mask:
            mono, dip = hp.fit_dipole(m_masked[i], gal_cut=gal_cut)            
        else:
            mono, dip = hp.fit_dipole(m_masked[i])
    
        # Subtract dipole map from data
        click.echo(click.style("Removing dipole:", fg="yellow"))
        click.echo(click.style("Dipole vector:",fg="green") + f" {dip}")
        click.echo(click.style("Dipole amplitude:",fg="green") + f" {np.sqrt(np.sum(dip ** 2))}")
        click.echo(click.style("Monopole:",fg="green") + f" {mono}")
        # Create dipole template
        ray = range(npix)
        vecs = hp.pix2vec(nside, ray)
        dipole = np.dot(dip, vecs)
        m[i] = m[i] - dipole - mono
    hp.write_map(input.replace(".fits", "_no-md.fits"), m, dtype=None, overwrite=True)


# fmt: off
@commands_fits.command()
@click.argument("input", type=click.STRING)
@click.argument("template", nargs=-1, type=click.STRING)
@click.option("-mask", type=click.STRING)
@click.option("-noise", type=click.STRING)
@click.option("-res", type=click.STRING)
# fmt: on
def fittemp(input, template, mask, noise, res):
    """
    Takes a fits map and fits the templates to it. 
    Outputs template fit values and uncertainty
    """
    field = (1,2)
    map_ = hp.read_map(input, field=(0,1), dtype=None, verbose=False)
    nside = hp.get_nside(map_)
    npix = hp.nside2npix(nside)
    ntemps = len(template)
    temp = np.zeros((ntemps, *map_.shape))

    for i in range(ntemps):
        temp[i] = hp.read_map(template[i], field=field, dtype=None, verbose=False)
    if mask:
        mask = np.logical_not(hp.read_map(mask, field=field, dtype=None, verbose=False))
    else:
        mask = np.logical_not(np.ones_like(map_))

    map_masked = hp.ma(map_)
    map_masked.mask = mask
    temp_masked = hp.ma(temp)
    temp_masked.mask = mask

    if noise: #noise
        with fits.open(noise) as hdulist:
            for hdu in hdulist:
                N = hdu.data
                N *= 1e-6
                print("scaling cov by 1e3")
    else:
        N = np.eye(2*npix)
    print("map:", map_.shape, "template:", temp.shape, "mask:", mask.shape, "N:", N.shape)
    t = temp_masked.reshape((ntemps, 2*npix)).T
    a = np.linalg.inv((t.T).dot(N).dot(t))
    print(a.shape)
    err = np.sqrt(np.diagonal(a))
    a = a.dot(t.T).dot(N).dot(map_masked.ravel())
    for i in range(len(a)):
        print(f"T_{i}: {a[i]:.3f} +- {err[i]:.5f}")
    residual = map_masked - np.sum(temp_masked*a.reshape(ntemps,1,1), axis=0)
    residual_masked = hp.ma(residual)
    residual_masked.mask = mask
    print("Sum of residual: ", np.sum(residual_masked))

    if res:
        hp.write_map(res, residual_masked, dtype=np.float32, overwrite=True)



from astropy.units import Quantity, Unit

from cosmoglobe.sky._base_components import DiffuseComponent
from cosmoglobe.sky.components._labels import SkyComponentLabel


class PowerLaw(DiffuseComponent):
    r"""Class representing the Synchrotron component in the sky model.

    Notes
    -----
    This is a generic power law given at a reference frequency
    :math:`\nu_{s,0}` with a power law :math:`\beta` in Rayleigh-Jeans
    temperature. It is defined using the convention in
    `BeyondPlanck (2020), Section 3.3.1
    <https://arxiv.org/pdf/2011.05609.pdf>`_;

    .. math::

        \boldsymbol{s}_\mathrm{RJ}^{\mathrm{s}}(\nu) \propto
        \left( \frac{\nu}{\nu_\mathrm{0,s}} \right)^
        {\beta + C \ln \nu / \nu_{0,s}},

    where :math:`\nu` is the frequency for which we are simulating the
    sky emission. :math:`C` is set to 0 for all current implementations
    as of BP9.
    """

    label = SkyComponentLabel.SYNCH
    freq_range = (0 * Unit("Hz"), 2 * Unit("THz"))

    def get_freq_scaling(self, freqs: Quantity, beta: Quantity) -> Quantity:
        """See base class."""

        return (freqs / self.freq_ref) ** beta



from typing import TYPE_CHECKING

import h5py
import numpy as np
from numba import njit

from cosmoglobe.h5 import PARAMETER_GROUP_NAME
from cosmoglobe.h5._exceptions import ChainKeyError

if TYPE_CHECKING:
    from cosmoglobe.h5.chain import Chain


def unpack_alms_from_chain(chain: "Chain", alms: np.ndarray, key: str) -> np.ndarray:
    """Unpacks alms from a chain."""

    root = key.split("/")[0]
    if root.startswith(PARAMETER_GROUP_NAME) or root in chain.samples:
        lmax_key = f"{key[:-3]}lmax"
    else:
        lmax_key = f"{chain.samples[0]}/{key[:-3]}lmax"

    try:
        with h5py.File(chain.path, "r") as file:
            lmax = file[lmax_key][()]
    except KeyError:
        raise ChainKeyError(
            f"{lmax_key} does not exist. Cannot unpack alms from chain "
            "without an lmax present in the chain"
        )

    if alms.ndim > 2:  # alms have shape (1, n) or (3, n)
        return np.asarray([unpack_alms(value, lmax) for value in alms])

    return unpack_alms(alms, lmax)


@njit
def unpack_alms(data, lmax):
    """Unpacks alms from the Commander chain output.

    Unpacking algorithm:
    https://github.com/trygvels/c3pp/blob/2a2937926c260cbce15e6d6d6e0e9d23b0be1262/src/tools.py#L9

    Parameters
    ----------
    data
        alms from a commander chainfile.
    lmax
        Maximum value for l used in the alms.

    Returns
    -------
    alms
        Unpacked version of the Commander alms (2-dimensional array)
    """

    n = len(data)
    n_alms = int(lmax * (2 * lmax + 1 - lmax) / 2 + lmax + 1)
    alms = np.zeros((n, n_alms), dtype=np.complex128)

    for sigma in range(n):
        i = 0
        for l in range(lmax + 1):
            j_real = l ** 2 + l
            alms[sigma, i] = complex(data[sigma, j_real], 0.0)
            i += 1

        for m in range(1, lmax + 1):
            for l in range(m, lmax + 1):
                j_real = l ** 2 + l + m
                j_comp = l ** 2 + l - m
                alms[sigma, i] = (
                    complex(
                        data[sigma, j_real],
                        data[sigma, j_comp],
                    )
                    / np.sqrt(2.0)
                )
                i += 1

    return alms



import time
import os
import numpy as np
import sys
import click
from cosmoglobe.release.tools import *

@click.group()
def commands():
    pass

@commands.command()
@click.argument("filename", type=click.STRING)
@click.argument("min", type=click.INT)
@click.argument("max", type=click.INT)
@click.argument("binfile", type=click.STRING)
def dlbin2dat(filename, min, max, binfile):
    """
    Outputs binned powerspectra averaged over a range of output samples.
    Filename Dl_[signal]_binned.dat.
    """
    signal = "cmb/Dl"

    import h5py

    dats = []
    with h5py.File(filename, "r") as f:
        for sample in range(min, max + 1):
            # Get sample number with leading zeros
            s = str(sample).zfill(6)

            # Get data from hdf
            data = f[s + "/" + signal][()]
            # Append sample to list
            dats.append(data)
    dats = np.array(dats)

    binned_data = {}
    possible_signals = ["TT","EE","BB","TE","EB","TB",]
    with open(binfile) as f:
        next(f)  # Skip first line
        for line in f.readlines():
            line = line.split()
            signal = line[0]
            if signal not in binned_data:
                binned_data[signal] = []
            signal_id = possible_signals.index(signal)
            lmin = int(line[1])
            lmax = int(line[2])
            ellcenter = lmin + (lmax - lmin) / 2
            # Saves (ellcenter, lmin, lmax, Dl_mean, Dl_stddev) over samples chosen
            binned_data[signal].append([ellcenter, lmin, lmax, np.mean(dats[:, signal_id, lmin], axis=0,), np.std(dats[:, signal_id, lmin], axis=0,),])

    header = f"{'l':22} {'lmin':24} {'lmax':24} {'Dl':24} {'stddev':24}"
    for signal in binned_data.keys():
        np.savetxt("Dl_" + signal + "_binned.dat", binned_data[signal], header=header,)


@commands.command()
@click.argument("label", type=click.STRING)
@click.argument("freqs", type=click.FLOAT, nargs=-1)
@click.argument("nside", type=click.INT,)
@click.option("-cmb", type=click.Path(exists=True), help="Include resampled chain file",)
@click.option("-synch", type=click.Path(exists=True), help="Include resampled chain file",)
@click.option("-dust", type=click.Path(exists=True), help="Include resampled chain file",)
@click.option("-ff", type=click.Path(exists=True), help="Include resampled chain file",)
@click.option("-ame", type=click.Path(exists=True), help="Include resampled chain file",)
#@click.option("-skipcopy", is_flag=True, help="Don't copy full .h5 file",)
def generate_sky(label, freqs, nside, cmb, synch, dust, ff, ame):
    """
    Generate sky maps from separate input maps.
    Reference frequencies from BP: CMB 1, SYNCH 30, DUST 545 353, FF 40, AME 22,
    Example:
    "c3pp generate-sky test 22 30 44 1024 -cmb BP_cmb_IQU_full_n1024_v1.0.fits -synch BP_synch_IQU_full_n1024_v1.0.fits"
    # Todo smoothing and nside?
    """
    import healpy as hp
    import numpy as np
    # Generate sky maps
    A = 1 # Only relative scaling
    for nu in freqs:
        filename = f"{label}_{nu}_n{nside}_generated.fits"
        print(f"Generating {filename}")
        data = np.zeros((3, hp.nside2npix(nside)))
        for pl in range(3):

            if cmb:
                pl_data  = hp.read_map(cmb, field=pl, verbose=False,)

            if synch:
                scaling  = fgs.lf(nu, A, betalf=-3.11, nuref=30.)
                print(hp.read_map(synch, field=pl, verbose=False)*scaling)
                pl_data += hp.read_map(synch, field=pl, verbose=False)*scaling

            if dust:
                if pl > 0:
                    scaling = fgs.dust(nu, A, beta=1.6, Td=18.5, nuref=353.,)
                else:
                    scaling = fgs.dust(nu, A, beta=1.6, Td=18.5, nuref=545.,)
                pl_data += hp.read_map(dust, field=pl, verbose=False,)*scaling

            if pl == 0:
                if ff:
                    scaling  = fgs.ff(nu, A, Te=7000., nuref=40.) 
                    pl_data += hp.read_map(ff, field=pl, verbose=False,)*scaling

                if ame:
                    scaling  = fgs.sdust(nu, A, nu_p=21, nuref=22.)
                    pl_data += hp.read_map(ame, field=pl, verbose=False,)*scaling

            data[pl] = pl_data
        hp.write_map(filename, data, dtype=None)



import time
import os
import numpy as np
import sys
import click
from cosmoglobe.release.tools import *

@click.group()
def commands_fits():
    pass

@commands_fits.command()
@click.argument("input", type=click.STRING)
def printheader(input,):
    """
    Prints the header of a fits file.
    """
    from astropy.io import fits

    with fits.open(input) as hdulist:
        hdulist.info()
        for hdu in hdulist:
            print(repr(hdu.header))

@commands_fits.command()
@click.argument("input", type=click.STRING)
def printdata(input,):
    """
    Prints the data of a fits file
    """
    from astropy.io import fits

    with fits.open(input) as hdulist:
        hdulist.info()
        for hdu in hdulist:
            print(repr(hdu.data))


@commands_fits.command()
@click.argument("input", type=click.STRING)
@click.argument("output", type=click.STRING)
@click.argument("columns", type=click.INT)
def rmcolumn(input, output, columns):
    """
    Removes columns in fits file
    """
    from astropy.io import fits

    with fits.open(input) as hdulist:
        for hdu in hdulist:
            hdu.header.pop(columns)        
            hdu.data.del_col(columns)        
        hdulist.writeto(output, overwrite=True)
        
@commands_fits.command()
@click.argument("input", type=click.STRING)
@click.option("-mask", type=click.STRING, help="Mask for dipole removal, 30deg sky cut by default")
@click.option("-sig", type=click.INT, multiple=True, help="fields to calculate, [0,1,2 by default]")
def rmmd(input, mask, sig):
    """
    removes the dipole and mask of input file
    if mask = "auto" uses 30 deg sky cut
    """
    import healpy as hp
    try:
        m = hp.read_map(input, field=(0,1,2), verbose=False, dtype=None,)
    except:
        m = hp.read_map(input, field=(0,), verbose=False, dtype=None,)
    nside = hp.get_nside(m)
    npix = hp.nside2npix(nside)
    print(f"Nside: {nside}, npix: {npix}")
    # Mask map for dipole estimation
    
    
    m_masked = hp.ma(m)
    if mask:
        m_masked.mask = np.logical_not(hp.read_map(mask,verbose=False,dtype=None,))
    
    # Fit dipole to masked map
    for i in sig:
        if not mask:
            mono, dip = hp.fit_dipole(m_masked[i], gal_cut=30)            
        else:
            mono, dip = hp.fit_dipole(m_masked[i])
    
        # Subtract dipole map from data
        click.echo(click.style("Removing dipole:", fg="yellow"))
        click.echo(click.style("Dipole vector:",fg="green") + f" {dip}")
        click.echo(click.style("Dipole amplitude:",fg="green") + f" {np.sqrt(np.sum(dip ** 2))}")
        click.echo(click.style("Monopole:",fg="green") + f" {mono}")
        # Create dipole template
        ray = range(npix)
        vecs = hp.pix2vec(nside, ray)
        dipole = np.dot(dip, vecs)
        m[i] = m[i] - dipole - mono
    hp.write_map(input.replace(".fits", "_no-md.fits"), m, dtype=None, overwrite=True)


@commands_fits.command()
@click.argument("input", type=click.STRING)
@click.argument("output", type=click.STRING)
def QU2ang(input, output,):
    """
    Calculates polarization angle map from QU signals.
    """
    import healpy as hp
    Q, U = hp.read_map(input, field=(1,2), dtype=None, verbose=False)
    phi = 0.5*np.arctan(U,Q)
    hp.write_map(output, phi, dtype=None, overwrite=True)


@commands_fits.command()
@click.argument("input", type=click.STRING)
@click.argument("template", nargs=-1, type=click.STRING)
@click.option("-mask", type=click.STRING)
@click.option("-noise", type=click.STRING)
@click.option("-res", type=click.STRING)
def fittemp(input, template, mask, noise, res):
    """
    Calculates polarization angle map from QU signals.
    """
    import healpy as hp
    field = (1,2)
    map_ = hp.read_map(input, field=(0,1), dtype=None, verbose=False)
    nside = hp.get_nside(map_)
    npix = hp.nside2npix(nside)
    ntemps = len(template)
    temp = np.zeros((ntemps, *map_.shape))
    #print("input: ", input, "template: ", template, "mask: ", mask)
    for i in range(ntemps):
        temp[i] = hp.read_map(template[i], field=field, dtype=None, verbose=False)
    if mask:
        mask = np.logical_not(hp.read_map(mask, field=field, dtype=None, verbose=False))
    else:
        mask = np.logical_not(np.ones_like(map_))

    map_masked = hp.ma(map_)
    map_masked.mask = mask
    temp_masked = hp.ma(temp)
    temp_masked.mask = mask

    if noise: #noise
        from astropy.io import fits
        with fits.open(noise) as hdulist:
            for hdu in hdulist:
                N = hdu.data
                N *= 1e-6
                print("scaling cov by 1e3")
    else:
        N = np.eye(2*npix)
    print("map:", map_.shape, "template:", temp.shape, "mask:", mask.shape, "N:", N.shape)
    t = temp_masked.reshape((ntemps, 2*npix)).T
    a = np.linalg.inv((t.T).dot(N).dot(t))
    print(a.shape)
    err = np.sqrt(np.diagonal(a))
    a = a.dot(t.T).dot(N).dot(map_masked.ravel())
    for i in range(len(a)):
        print(f"T_{i}: {a[i]:.3f} +- {err[i]:.5f}")
    residual = map_masked - np.sum(temp_masked*a.reshape(ntemps,1,1), axis=0)
    residual_masked = hp.ma(residual)
    residual_masked.mask = mask
    print("Sum of residual: ", np.sum(residual_masked))
    if res:
        hp.write_map(res, residual_masked, dtype=np.float32, overwrite=True)




@commands_fits.command()
@click.argument("input1", type=click.STRING)
@click.argument("input2", type=click.STRING)
@click.argument("output", type=click.STRING)
@click.option("-beam1", type=click.STRING, help="Optional beam file for input 1",)
@click.option("-beam2", type=click.STRING, help="Optional beam file for input 2",)
@click.option("-mask", type=click.STRING, help="Mask",)
def crosspec(input1, input2, output, beam1, beam2, mask,):
    """
    Calculates a powerspectrum from polspice.
    Using this path /mn/stornext/u3/trygvels/PolSpice_v03-03-02/
    """
    sys.path.append("/mn/stornext/u3/trygvels/PolSpice_v03-03-02/")
    from ispice import ispice

    lmax = 6000
    fwhm = 0
    #mask = "dx12_v3_common_mask_pol_005a_2048_v2.fits"
    if beam1 and beam2:
        ispice(input1,
               clout=output,
               nlmax=lmax,
               beam_file1=beam1,
               beam_file2=beam2,
               mapfile2=input2,
               maskfile1=mask,
               maskfile2=mask,
               fits_out="NO",
               polarization="YES",
               subav="YES",
               subdipole="YES",
               symmetric_cl="YES",
           )
    else:

        ispice(input1,
               clout=output,
               nlmax=lmax,
               beam1=0.0,
               beam2=0.0,
               mapfile2=input2,
               maskfile1=mask,
               maskfile2=mask,
               fits_out="NO",
               polarization="YES",
               subav="YES",
               subdipole="YES",
               symmetric_cl="YES",
           )

@commands_fits.command()
@click.argument("input", type=click.STRING)
@click.argument("output", type=click.STRING)
@click.option("-min", default=1, type=click.INT, help="Start sample, default 1",)
@click.option("-max", default=None, type=click.INT, help="End sample, calculated automatically if not set",)
@click.option("-minchain", default=1, help="lowest chain number, c0002 [ex. 2] (default=1)",)
@click.option("-maxchain", default=1, help="max number of chains c0005 [ex. 5] (default=1)",)
@click.option("-chaindir", default=None, type=click.STRING, help="Base of chain directory, overwrites chain iteration from input file name to iteration over chain directories, BP_chain_c15 to BP_chain_c19 [ex. 'BP_chain', with minchain = 15 and maxchain = 19]",)
@click.option("-fwhm", default=0.0, help="FWHM in arcmin")
@click.option("-nside", default=None, type=click.INT, help="Nside for down-grading maps before calculation",)
@click.option("-zerospin", is_flag=True, help="If smoothing, treat maps as zero-spin maps.",)
@click.option("-missing", is_flag=True, help="If files are missing, drop them. Else, exit computation",)
@click.option("-pixweight", default=None, type=click.STRING, help="Path to healpy pixel weights.",)
def fits_mean(
        input, output, min, max, minchain, maxchain, chaindir, fwhm, nside, zerospin, missing, pixweight):
    """
    Calculates the mean over sample range from fits-files.
    ex. res_030_c0001_k000001.fits res_030_20-100_mean_40arcmin.fits -min 20 -max 100 -fwhm 40 -maxchain 3\n
    If output name is set to .dat, data will not be converted to map.

    Note: the input file name must have the 'c0001' chain identifier and the 'k000001' sample identifier. The -min/-max and -chainmin/-chainmax options set the actual samples/chains to be used in the calculation 
    """

    fits_handler(input, min, max, minchain, maxchain, chaindir, output, fwhm, nside, zerospin, missing, pixweight, np.mean, write=True)

@commands_fits.command()
@click.argument("input", type=click.STRING)
@click.argument("output", type=click.STRING)
@click.option("-min", default=1, type=click.INT, help="Start sample, default 1",)
@click.option("-max", default=None, type=click.INT, help="End sample, calculated automatically if not set",)
@click.option("-minchain", default=1, help="lowest chain number, c0002 [ex. 2] (default=1)",)
@click.option("-maxchain", default=1, help="max number of chains c0005 [ex. 5] (default=1)",)
@click.option("-chaindir", default=None,type=click.STRING, help="Base of chain directory, overwrites chain iteration from input file name to iteration over chain directories, BP_chain_c15 to BP_chain_c19 [ex. 'BP_chain', with minchain = 15 and maxchain = 19]",)
@click.option("-fwhm", default=0.0, help="FWHM in arcmin")
@click.option("-nside", default=None, type=click.INT, help="Nside for down-grading maps before calculation",)
@click.option("-zerospin", is_flag=True, help="If smoothing, treat maps as zero-spin maps.",)
@click.option("-missing", is_flag=True, help="If files are missing, drop them. Else, exit computation",)
@click.option("-pixweight", default=None, type=click.STRING, help="Path to healpy pixel weights.",)
def fits_stddev(
        input, output, min, max, minchain, maxchain, chaindir, fwhm, nside, zerospin, missing, pixweight):
    """
    Calculates the standard deviation over sample range from fits-files.
    ex. res_030_c0001_k000001.fits res_030_20-100_mean_40arcmin.fits -min 20 -max 100 -fwhm 40 -maxchain 3
    If output name is set to .dat, data will not be converted to map.

    Note: the input file name must have the 'c0001' chain identifier and the 'k000001' sample identifier. The -min/-max and -chainmin/-chainmax options set the actual samples/chains to be used in the calculation 
    """

    fits_handler(input, min, max, minchain, maxchain, chaindir, output, fwhm, nside, zerospin, missing, pixweight, np.std, write=True)



# This file contains useful functions used across different plotting scripts.
from cmath import inf
from re import A, T
import warnings

from numpy.core.numeric import NaN
from .. import data as data_dir
from cosmoglobe.sky.model import SkyModel
from cosmoglobe.h5.chain import Chain
from cosmoglobe.sky._units import cmb_equivalencies, Unit

import cmasher
from rich import print
import numpy as np
import healpy as hp
import astropy.units as u
import matplotlib.pyplot as plt
import matplotlib as mpl
from matplotlib.colors import colorConverter, LinearSegmentedColormap, ListedColormap
from matplotlib.patches import Polygon
from matplotlib import _pylab_helpers
from pathlib import Path
import json

DEFAULT_FONTSIZES = {
    "xlabel": 11,
    "ylabel": 11,
    "xtick_label": 8,
    "ytick_label": 8,
    "title": 12,
    "cbar_label": 11,
    "cbar_tick_label": 11,
    "llabel": 11,
    "rlabel": 11,
}

STOKES = [
    "I",
    "Q",
    "U",
]


def set_style(darkmode=False, font="serif"):
    """
    This function sets the color parameter and text style
    """
    if font == "serif":
        plt.rc(
            "font",
            family="serif",
        )
        plt.rcParams["mathtext.fontset"] = "stix"
    plt.rc(
        "text.latex",
        preamble=r"\usepackage{sfmath}",
    )

    tex_fonts = {
        # Use 10pt font in plots, to match 10pt font in document
        "axes.labelsize": 11,
        "font.size": 11,
        # Make the legend/label fonts a little smaller
        "legend.fontsize": 8,
        "xtick.labelsize": 11,
        "ytick.labelsize": 11,
    }
    plt.rcParams.update(
        tex_fonts,
    )

    ## If darkmode plot (white frames), change colors:
    params = [
        "text.color",
        "axes.edgecolor",
        "axes.labelcolor",
        "xtick.color",
        "ytick.color",
        "grid.color",
        "legend.facecolor",
        "legend.edgecolor",
    ]
    if darkmode:
        for p in params:
            plt.rcParams[p] = "white"
    else:
        for p in params:
            plt.rcParams[p] = "black"


def get_figure_width(width=600, fraction=1):
    """Set figure dimensions to avoid scaling in LaTeX.

    Parameters
    ----------
    width: float
            Document textwidth or columnwidth in pts
    fraction: float, optional
            Fraction of the width which you wish the figure to occupy

    Returns
    -------
    fig_dim: tuple
            Dimensions of figure in inches
    """
    # Width of figure (in pts)
    fig_width_pt = width * fraction

    # Convert from pt to inches
    inches_per_pt = 1 / 72.27

    # Golden ratio to set aesthetic figure height
    # https://disq.us/p/2940ij3
    golden_ratio = (5**0.5 - 1) / 2

    # Figure width in inches
    fig_width_in = fig_width_pt * inches_per_pt
    # Figure height in inches
    fig_height_in = fig_width_in * golden_ratio

    # Proper scaling
    fig_dim = (fig_width_in, fig_height_in)
    #print(f"Using dimensions: {fig_dim}")
    return fig_dim


def make_fig(figsize, fignum, hold, sub, reuse_axes, darkmode=False, projection=None, fraction=0.5):
    """
    Create matplotlib figure, add subplot, use current axes etc.
    """

    if figsize is None:
        figsize = get_figure_width(fraction=fraction)

    #  From healpy
    nrows, ncols, idx = (1, 1, 1)
    width, height = figsize
    if not (hold or sub or reuse_axes):
        # Set general style parameters
        set_style(darkmode)

        fig = plt.figure(fignum, figsize=(width, height))
    elif hold:
        fig = plt.gcf()
        # left, bottom, right, top = np.array(fig.gca().get_position()).ravel()
        fig.delaxes(fig.gca())
    elif reuse_axes:
        fig = plt.gcf()
    else:  # using subplot syntax
        if hasattr(sub, "__len__"):
            nrows, ncols, idx = sub
        else:
            nrows, ncols, idx = sub // 100, (sub % 100) // 10, (sub % 10)

        # If figure exists, get. If not, create.
        figManager = _pylab_helpers.Gcf.get_active()
        if figManager is not None:
            fig = figManager.canvas.figure
        else:
            # Set general style parameters
            set_style(darkmode)
            figsize = (width * ncols, height * nrows)
            fig = plt.figure(figsize=figsize)

        if idx < 1 or idx > ncols * nrows:
            raise ValueError("Wrong values for subplot: %d, %d, %d" % (nrows, ncols, idx))

    if reuse_axes:
        ax = fig.gca()
    else:
        ax = fig.add_subplot(nrows,ncols,idx, projection=projection)

    return fig, ax


def get_percentile(m, percentile):
    """
    This function gets appropriate min and max
    values of the data for a given percentile
    """
    vmin = np.percentile(m[~m.mask], 100.0 - percentile)
    vmax = np.percentile(m[~m.mask], percentile)

    vmin = 0.0 if abs(vmin) < 1e-5 else vmin
    vmax = 0.0 if abs(vmax) < 1e-5 else vmax
    return [vmin, vmax]


def fmt(x, pos):
    """
    Format color bar labels
    """
    if abs(x) >= 1e4 or (abs(x) <= 1e-3 and abs(x) > 0):
        a, b = f"{x:.2e}".split("e")
        b = int(b)
        if float(a) == 1.00:
            return r"$10^{" + str(b) + "}$"
        elif float(a) == -1.00:
            return r"$-10^{" + str(b) + "}$"
        else:
            return rf"${a} \cdot 10^{b}$"
    elif abs(x) > 1e1 or (float(abs(x))).is_integer():
        return rf"${int(x):d}$"
    else:
        x = round(x, 2)
        return rf"${x}$"


def format_list(items):
    return [fmt(i, 1) for i in items]


def symlog(m, linthresh=1.0):
    """
    This is the semi-logarithmic function used when logscale=True
    """
    if linthresh == 0.0:
        return np.sign(m) * np.log10(1e-20 + abs(m))

    # Extra fact of 2 ln 10 makes symlog(m) = m in linear regime
    m = m / linthresh / (2 * np.log(10))
    return np.log10(0.5 * (m + np.sqrt(4.0 + m * m)))


def legend_positions(input):
    """
    Calculate position of labels to the right in plot...
    """
    positions = input[-1, :]

    def push(dpush):
        """
        ...by puting them to the last y value and
        pushing until no overlap
        """
        collisions = 0
        for column1, value1 in enumerate(positions):
            for column2, value2 in enumerate(positions):
                if column1 != column2:
                    dist = abs(value1 - value2)
                    if dist < dpush * 10:
                        collisions += 1
                        if value1 < value2:
                            positions[column1] -= dpush
                            positions[column2] += dpush
                        else:
                            positions[column1] += dpush
                            positions[column2] -= dpush
                            return True

    min_ = np.min(np.min(input))
    max_ = np.max(np.max(input))
    dpush = np.abs(max_ - min_) * 0.01

    pushings = 0
    while True:
        if pushings == 1000:
            dpush *= 10
            pushings = 0
        pushed = push(dpush)
        if not pushed:
            break
        if pushings > 100:
            break
        pushings += 1
    return positions


def gradient_fill(x, y, fill_color=None, ax=None, alpha=1.0, invert=False, **kwargs):
    """
    Plot a line with a linear alpha gradient filled beneath it.

    Parameters
    ----------
    x, y : array-like
        The data values of the line.
    fill_color : a matplotlib color specifier (string, tuple) or None
        The color for the fill. If None, the color of the line will be used.
    ax : a matplotlib Axes instance
        The axes to plot on. If None, the current pyplot axes will be used.
    Additional arguments are passed on to matplotlib's ``plot`` function.

    Returns
    -------
    line : a Line2D instance
        The line plotted.
    im : an AxesImage instance
        The transparent gradient clipped to just the area beneath the curve.
    """

    if ax is None:
        ax = plt.gca()

    (line,) = ax.plot(x, y, **kwargs)
    if fill_color is None:
        fill_color = line.get_color()

    zorder = line.get_zorder()
    # alpha = line.get_alpha()
    alpha = 1.0 if alpha is None else alpha

    z = np.empty((100, 1, 4), dtype=float)
    rgb = colorConverter.to_rgb(fill_color)
    z[:, :, :3] = rgb
    z[:, :, -1] = np.linspace(0, alpha, 100)[:, None]

    xmin, xmax, ymin, ymax = x.min(), x.max(), y.min(), y.max()
    if invert:
        ymin, ymax = (ymax, ymin)
    im = ax.imshow(z, aspect="auto", extent=[xmin, xmax, ymin, ymax], origin="lower", zorder=zorder)

    xy = np.column_stack([x, y])
    xy = np.vstack([[xmin, ymin], xy, [xmax, ymin], [xmin, ymin]])
    clip_path = Polygon(xy, facecolor="none", edgecolor="none", closed=True)
    ax.add_patch(clip_path)
    im.set_clip_path(clip_path)

    ax.autoscale(True)
    return line, im


def gradient_fill_between(ax, x, y1, y2, color="#ffa15a", alpha=0.8):
    N = 100
    y = np.zeros((N, len(y1)))
    for i in range(len(y1)):
        y[:, i] = np.linspace(y1[i], y2[i], N)

    alpha = np.linspace(0, alpha, 100) ** 1.5
    for i in range(1, N):
        ax.fill_between(x, y[i - 1], y[i], color=color, alpha=alpha[i], zorder=-10)
        ax.set_rasterization_zorder(-1)


def find_nearest(array, value):
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return array[idx], idx


def apply_logscale(m, ticks, linthresh=1):
    """
    This function converts the data to logscale,
    and also converts the tick locations correspondingly
    NOTE: This should ideally be done for the colorbar, not the data.
    """

    print(f"[magenta]Applying semi-logscale with linear threshold={linthresh}[/magenta]")
    m = symlog(m, linthresh)
    new_ticks = []
    for i in ticks:
        new_ticks.append(symlog(i, linthresh))

    m = np.maximum(np.minimum(m, new_ticks[-1]), new_ticks[0])
    return m, new_ticks


def standalone_colorbar(
    cmap,
    ticks,
    ticklabels=None,
    unit=None,
    fontsize=None,
    norm=None,
    shrink=0.3,
    darkmode=False,
    width=2,
    extend=None,
):

    fig = plt.figure(figsize=(width, width / 4))
    set_style(darkmode)
    cmap = load_cmap(cmap)
    img = plt.imshow(np.array([[0, 1]]), cmap=cmap)
    plt.gca().set_visible(False)
    ax = plt.axes([0.1, 0.7, 0.8, 0.2])  # LBWH

    if ticklabels is None:
        ticklabels = format_list(ticks)

    apply_colorbar(
        fig,
        ax,
        None,
        ticks,
        ticklabels,
        unit,
        fontsize=fontsize,
        norm=norm,
        cbar_shrink=shrink,
        cbar_pad=0.0,
        cmap=cmap,
        extend=extend
    )


def apply_colorbar(
    fig,
    ax,
    image,
    ticks,
    ticklabels,
    unit,
    fontsize=None,
    linthresh=1,
    norm=None,
    cbar_pad=0.04,
    cbar_shrink=0.3,
    cmap=None,
    orientation="horizontal",
    extend=None
):
    """
    This function applies a colorbar to the figure and formats the ticks.
    """
    if isinstance(unit, u.UnitBase):
        unit = unit.to_string("latex")

    if fontsize is None:
        fontsize = DEFAULT_FONTSIZES#['cbar_label']
    #elif type(fontsize) is dict:
    #    fontsize = fontsize['cbar_label']

    if image is None and cmap is not None:
        norm = mpl.colors.Normalize(vmin=ticks[0], vmax=ticks[-1])
        cb = mpl.colorbar.ColorbarBase(
            ax,
            cmap=cmap,
            norm=norm,
            orientation=orientation,
            ticks=ticks,
            extend=extend,
        )
        cb.set_label(label=unit, size=fontsize['cbar_label'])
    else:
        cb = fig.colorbar(
            image,
            # ax=ax,
            orientation=orientation,
            shrink=cbar_shrink,
            pad=cbar_pad,
        )
    if orientation == 'horizontal':
       cb.ax.set_xticklabels(ticklabels, size=fontsize["cbar_tick_label"])
       cb.ax.xaxis.set_label_text(unit, size=fontsize["cbar_label"])
    elif orientation == 'vertical':
       cb.ax.set_yticklabels(ticklabels, size=fontsize["cbar_tick_label"])
       cb.ax.yaxis.set_label_text(unit, size=fontsize["cbar_label"])


    if norm in ["linlog", "log"]:
        """
        Make logarithmic tick markers manually
        """
        linticks = np.linspace(-1, 1, 3) * linthresh if linthresh > 0.0 else [0]
        logmin = np.round(ticks[0])
        logmax = np.round(ticks[-1])
        logticks_min = -(10 ** np.arange(0, abs(logmin) + 1))
        logticks_max = 10 ** np.arange(0, logmax + 1)
        ticks_ = np.unique(np.concatenate((logticks_min, linticks, logticks_max)))

        # cb.set_ticks(np.concatenate((ticks,symlog(ticks_))), []) # Set major ticks
        logticks = symlog(ticks_, linthresh)
        logticks = [x for x in logticks if x not in ticks]
        if orientation == "horizontal":
            cb.set_xticks(np.concatenate((ticks, logticks)))  # Set major ticks
            cb.ax.set_xticklabels(ticklabels + [""] * len(logticks), ha='center')
        elif orientation == "vertical":
            cb.ax.yaxis.set_ticks(np.concatenate((ticks, logticks)))  # Set major ticks
            cb.ax.set_yticklabels(ticklabels + [""] * len(logticks))

        # Minor tick log markers

        # Linear range for linlog plotting
        minorticks = np.linspace(-linthresh, linthresh, 5) if linthresh > 0.0 else [0]

        # Make first range of logarithmic minor ticks
        lt = 1 if norm == "log" else linthresh
        minorticks2 = np.arange(2, 10) * lt

        # Create range of minor ticks from min value to negative max
        for i in range(len(logticks_min)):
            minorticks = np.concatenate((-(10**i) * minorticks2, minorticks))

        # Create range of minor ticks from min value to positive max
        for i in range(len(logticks_max)):
            minorticks = np.concatenate((minorticks, (10**i) * minorticks2))

        # Convert values to true logarithmic values
        minorticks = symlog(minorticks, linthresh)

        # Remove values outside of range
        minorticks = minorticks[(minorticks >= ticks[0]) & (minorticks <= ticks[-1])]
        if orientation == "horizontal":
            cb.ax.xaxis.set_ticks(minorticks, minor=True)
        elif orientation == "vertical":
            cb.ax.yaxis.set_ticks(minorticks, minor=True)
    # workaround for issue with viewers, see colorbar docstring
    cb.solids.set_edgecolor("face")
    bbox = ax.get_window_extent().transformed(fig.dpi_scale_trans.inverted())
    bwidth, bheight = bbox.width, bbox.height
    points_per_inch = 72
    if orientation == "horizontal":
        cb.ax.tick_params(
            which="both",
            axis="x",
            direction="in",
            color="#3d3d3d",
            labelsize=fontsize["cbar_tick_label"],
            width=fontsize['cbar_tick_label']/10,
            #length=bheight*points_per_inch*0.75,
            length=bheight*points_per_inch*0.02,
        )
        cb.ax.xaxis.labelpad = 0
        if norm in ["linlog", "log"]:
            ticklabels = cb.ax.get_xticklabels()
        cb.ax.set_xticks(ticks)
        cb.ax.set_xticklabels(ticklabels)
    elif orientation == "vertical":
        cb.ax.tick_params(
            which="both",
            axis="y",
            direction="in",
            color="#3d3d3d",
            labelsize=fontsize["cbar_tick_label"],
            width=fontsize['cbar_tick_label']/10,
            length=bheight*points_per_inch*0.02,
        )
        cb.ax.yaxis.labelpad = 0

        if norm in ["linlog", "log"]:
            ticklabels = cb.ax.get_yticklabels()
        cb.ax.set_yticks(ticks)
        cb.ax.set_yticklabels(ticklabels, Rotation=90)

    return  # cb


def load_cmap(cmap):
    """
    This function takes the colormap label and loads the colormap object
    """

    if cmap is None:
        cmap = "planck"

    if cmap in ["planck", "planck_log", "wmap"]:
        cmap_path =  cmap_path = Path(data_dir.__path__[0]) / f"{cmap}_cmap.dat"
        cmap = ListedColormap(np.loadtxt(cmap_path) / 255.0, cmap)
    elif cmap.startswith("black2"):
        cmap = LinearSegmentedColormap.from_list(cmap, cmap.split("2"))
    else:
        try:
            cmap = eval(f"cmasher.{cmap}")
        except:
            try:

                cmap = eval(f"cm.{cmap}")
            except:

                cmap = plt.get_cmap(cmap)

    return cmap


def get_data(input, sig, comp, freq, fwhm, nside=None, sample=-1, scale=1,
    remove_mono=None, remove_dip=None, nest=False, gal_cut=0):
    # Parsing component string
    fwhm_ = None
    specparam = "amp"
    if comp is not None:
        comp, *specparam_ = comp.split()
        if specparam_:
            specparam = specparam_[0]

    # If map is string, read data
    if isinstance(input, str):
        # If string input, then read either hdf or fits
        if input.endswith(".h5"):
            if comp is None:
                if freq is None:
                    print("[bold magenta]Warning! Neither frequency nor component selected. Plotting sky at 70GHz[/bold magenta]")
                data = SkyModel.from_chain(input, components=comp, nside=nside, samples=sample)
            else:
                if freq is not None:
                    # If frequency is specified, simulate sky with model.
                    data = SkyModel.from_chain(input, components=comp, nside=nside, samples=sample)
                else:
                    chain = Chain(
                        input,
                    )
                    lmax = chain.get(f"{comp}/{specparam}_lmax", samples=sample)
                    alms = chain.get(f"{comp}/{specparam}_alm", samples=sample)
                    fwhm_ = chain.parameters[comp]["fwhm"] * u.arcmin
                    if nside is None:
                        nside = chain.parameters[comp]["nside"]
                    pol = True if specparam == "amp" and alms.shape[0] == 3 else False
                    # PROBLEM! Get issue with nside when udgrading directly to low nside
                    data = hp.alm2map(
                        alms,
                        nside=nside,
                        fwhm=fwhm_.to(u.rad).value,
                        lmax=lmax,
                        mmax=lmax,
                        pol=pol,
                    )

        elif input.endswith(".fits"):
            data = hp.read_map(input, field=sig)
        else:
            raise ValueError("Input file must be .fits or .h5")
    else:
        data = input

    if isinstance(data, SkyModel):
        """
        Get data from model object with frequency scaling
        """
        if comp is None:
            if freq is None:
                print("[bold magenta]Warning! Neither frequency nor component selected. Plotting sky at 70GHz[/bold magenta]")
                freq = 70 * u.GHz
            comp = "freqmap"
            m = data(freq, fwhm=fwhm)
        else:
            if specparam == "amp":
                if freq is None:
                    # Get map at reference frequency if freq not specified
                    freq_ref = data[comp].freq_ref
                    freq = (freq_ref.value).squeeze()
                    m = data[comp].amp
                    if comp == "radio":
                        m = data(freq_ref, fwhm=fwhm, components=[comp])
                    
                    try:
                        freq = np.round(freq[sig], 5) * freq_ref.unit
                    except IndexError:
                        freq = np.round(freq, 5) * freq_ref.unit
                else:
                    # If freq is specified, scale with sky model
                    m = data(freq, fwhm=fwhm, components=[comp])
            else:
                m = data[comp].spectral_parameters[specparam]

                if len(m[sig]) == 1:
                    warnings.warn("Same value across the whole sky, mapping to array of length Npix")
                    m = np.full(hp.nside2npix(data.nside), m[sig])

    elif isinstance(data, np.ndarray):
        m = data
        if m.ndim == 1 and sig > 0 and not isinstance(input, str):
            print(f"[magenta]Input array is 1d, specified signal will only be used for labeling, make sure this is correct.[/magenta]")
    else:
        raise TypeError(f"Type {type(data)} of data not supported" f"Supports numpy array, cosmoglobe model object or fits file string")

    # Make sure it is a 1d array
    if m.ndim > 1:
        if sig>np.shape(m)[0]-1:
            warnings.warn(f"Index {sig} out of data range {np.shape(m)}, returning empty map")
            m = np.zeros(np.shape(m[-1]))+hp.UNSEEN
        else:
            m = m[sig]

    m = hp.ma(m)

    # Convert astropy map to numpy array
    if isinstance(m, u.Quantity):
        m = m.value

    # ud_grade map
    if nside is not None and nside != hp.get_nside(m):
        print("ud_grading map")
        m = hp.ud_grade(m, nside)

    # Smooth map
    if fwhm > 0.0 and fwhm_ is None:
        m = hp.smoothing(m, fwhm.to(u.rad).value)

    # Rescale map
    m = m * scale

    # Remove monopole and dipole
    if remove_dip:
        if remove_dip == 'auto':
           gal_cut = 30
        m = hp.remove_dipole(m, gal_cut=gal_cut, nest=nest, copy=True)
    elif remove_mono:
        if remove_mono == 'auto':
           gal_cut = 30
        m = hp.remove_monopole(m, gal_cut=gal_cut, nest=nest, copy=True)


    return m, comp, freq, hp.get_nside(m)


def get_params(**params):
    """
    This parses the autoparams json file for automatically setting parameters
    for an identified sky component.
    """
    freq = params["freq_ref"]
    # Convert astropy unit to latex
    if isinstance(params["unit"], u.UnitBase):
        params["unit"] = params["unit"].to_string("latex")

    # Autoset plotting parameters if component is specified
    if params["comp"] is not None:
        with open(Path(data_dir.__path__[0]) / "autoparams.json", "r") as f:
            autoparams = json.load(f)
        try:
            autoparams = autoparams[params["comp"]]
        except KeyError:
            print(f"[magenta]Component label {params['comp']} not found. Using unidentified profile.[/magenta]")
            print(f"available keys are: {autoparams.keys()}")
            autoparams = autoparams["unidentified"]

        # Replace default values if None
        params = {key: (autoparams[key] if key in autoparams and value is None else value) for key, value in params.items()}

        # NOTE: This is really ugly code for selecting pol parameters
        # Pick polarization or temperature
        l = 0 if params["sig"] < 1 else -1
        for key in ["cmap", "ticks", "freq_ref"]:
            if isinstance(params[key], list):
                if key == "ticks":
                    if isinstance(params[key][0], list):
                        params[key] = params[key][l]
                else:
                    params[key] = params[key][l]

        if params["freq_ref"] is not None:
            params["freq_ref"] * u.GHz

        if params["llabel"] is None:
            params["llabel"] = ["I", "Q", "U"][params["sig"]]
        specials = ["residual", "freqmap", "bpcorr", "smap"]
        if any(j in params["comp"] for j in specials):
            if freq is not None:
                params["rlabel"] = params["rlabel"] + "{" + f'{("%.5f" % freq.value).rstrip("0").rstrip(".")}' + "}"
            else:
                params["rlabel"] = None
                warnings.warn(f'Specify frequency with -freq for automatic frequency labeling with the "freqmap" profile')

        """
        # NOTE: This probably doesnt work with the current state of the comp
        # keyword. Rewrite somehow.
        if "rms" in params["comp"]:
            params["rlabel"] += "^{\mathrm{RMS}}"
            params["cmap"] = "neutral"
            params["ticks"] = "comp"
        if "stddev" in params["comp"]:
            params["rlabel"] = "\sigma_{\mathrm{" + params["rlabel"] + "}}"
            params["cmap"] = "neutral"
            params["ticks"] = "comp"
        if "mean" in params["comp"]:
            params["rlabel"] = "\langle " + params["rlabel"] + "\rangle"
        """

        # Specify at which frequency we are observing in unit lable
        if freq is not None and params["unit"] is not None:
            params["unit"] = f'{params["unit"]}\,@\,{("%.5f" % freq.value).rstrip("0").rstrip(".")}' + "\,\mathrm{GHz}"

        # In the case that a difference frequency than the reference
        # is requested for a component map.
        if params["ticks"] == None:
            if freq is not None and params["freq_ref"] != freq.value and params["comp"] not in specials:
                warnings.warn(f"Input frequency is different from reference, autosetting ticks")
                print(f'input: {freq}, reference: {params["freq_ref"]}')
                params["ticks"] = "auto"

    # If ticks is not set automatically, try inputs
    if params["ticks"] is None:
        params["ticks"] = [params["min"], params["max"]]

    # if ticks is not specified to be auto in function call
    if params["ticks"] != "auto":
        # If range is not specified in function call
        if params["rng"] is not None:
            params["ticks"] = [-params["rng"], 0.0, params["rng"]]
        else:
            # If min is not specified in function call
            if params["min"] is not None:
                # If ticks is set to be automatic from autoparams
                if params["ticks"] == "auto":
                    params["ticks"] = [None, None]  # (These will be set later)
                params["ticks"][0] = params["min"]
            # If max is not specified in function call
            if params["max"] is not None:
                # If tick is set to be automatic from atoparams
                if params["ticks"] == "auto":
                    params["ticks"] = [None, None]  # (These will be set later)
                params["ticks"][-1] = params["max"]

    # Ticks and ticklabels
    if params["ticks"] == "auto" or "auto" in params["ticks"]:
        params["ticks"] = get_percentile(params["data"], 99)
    elif None in params["ticks"]:
        pmin, pmax = get_percentile(params["data"], 99)
        if params["ticks"][0] is None:
            params["ticks"][0] = pmin
        if params["ticks"][-1] is None:
            params["ticks"][-1] = pmax

    # Math text in labels
    for key in ["rlabel", "llabel", "unit"]:
        if params[key] and params[key] != "":
            if "$" not in params[key]:
                params[key] = r"$" + params[key] + "$"

    # Special case if dipole is detected in freqmap
    if params["comp"] == "freqmap" and params["min"] is None and params["max"] is None and params["rng"] is None:
        # if colorbar is super-saturated
        while len(params["data"][abs(params["data"]) > params["ticks"][-1]]) / len(params["data"].ravel()) > 0.7:
            params["ticks"] = [tick * 10 for tick in params["ticks"]]
            print("[magenta]Colormap saturated. Expanding color-range.[/magenta]")

    # Create ticklabels from final ticks
    params["ticklabels"] = format_list(params["ticks"])

    # Set planck as default cmap
    if params["cmap"] is None:
        params["cmap"] = "planck"
    
    return params


def mask_map(m, mask):
    """
    Apply a mask to a map
    """

    m = hp.ma(m)
    m.mask = mask
    # Alternative

    # mask = mask*np.NaN
    # m *= mask
    return m


def create_70GHz_mask(sky_frac, nside=256, pol=False):
    """
    Creates a mask from a 70GHz frequency map at nside=256 for a
    threshold corresponding to a given sky fraction (in %). The 70GGz
    map is chosen due to it containing a large portion of all low-frequency
    sky components.
    Parameters
    ----------
    sky_frac : float
        Sky fraction in percentage.
    Returns
    -------
    mask : numpy.ndarray
        Mask covering the sky for a given sky fraction.
    """
    # Read amplitude map to use for thresholding
    field = (1, 2) if pol else 0
    template = hp.read_map(
        Path(data_dir.__path__[0]) / "mask_template_n256.fits",
        dtype=np.float64,
        field=field,
    )
    if pol:
        template = np.sqrt(template[0] ** 2 + template[1] ** 2)
    template = hp.ma(template)

    if nside != 256:
        template = hp.ud_grade(template, nside)

    # Masking based on sky fraction is not trivial. Here we manually compute
    # the sky fraction by masking all pixels with amplitudes larger than a
    # given percentage of the maximum map amplitude. The pixels masks then
    # correspond to a sky fraction. We tabulate the amplitude percentage and
    # sky fraction for a range, and interpolate from this table.
    amp_percentages = np.flip(np.arange(1, 101))
    fracs = []
    mask = np.zeros(len(template), dtype=bool)

    for i in range(len(amp_percentages)):
        mask = np.zeros(len(template), dtype=bool)
        masked_template = np.abs(hp.ma(template))
        mask[np.where(np.log(masked_template) > (amp_percentages[i] / 100) * np.nanmax(np.log(masked_template)))] = 1
        masked_template.mask = mask

        frac = ((len(masked_template) - masked_template.count()) / len(masked_template)) * 100
        fracs.append(frac)

    amp_percentage = np.interp(100 - sky_frac, fracs, amp_percentages)

    mask = np.zeros(len(template), dtype=bool)
    masked_template = np.abs(hp.ma(template))
    mask[np.where(np.log(masked_template) > (amp_percentage / 100) * np.nanmax(np.log(masked_template)))] = 1

    return mask


def seds_from_model(nu, model, nside=None, pol=False, sky_fractions=(25, 85)):
    ignore_comps = ["radio"]
    comps = model.components
    if nside is None or nside == model.nside:
        nside = model.nside
        udgrade = False
    else:
        udgrade = True

    masks = np.zeros((len(sky_fractions), hp.nside2npix(nside)))

    for i, sky_frac in enumerate(sky_fractions):
        masks[i] = create_70GHz_mask(sky_frac, nside, pol=pol)

    # SED dictionary with T and P, skyfractions and the value per nu
    seds = {comp: np.zeros((2, len(sky_fractions), len(nu))) for comp in model.components}
    for key, value in comps.items():
        if key in ignore_comps:
            continue

        for i, mask in enumerate(masks):
            specinds = {}
            for specind, m in value.spectral_parameters.items():
                if m.shape[-1] != 1:
                    m = np.mean(mask_map(m, mask), axis=1).reshape(-1, 1) * value.spectral_parameters[specind].unit
                specinds[specind] = m

            amp = hp.ud_grade(value.amp, nside) if udgrade else value.amp
            amp[0, amp[0] < 0.0] = 0.0  # abs(amp)
            amp = mask_map(amp, mask)

            amp_pol = 0
            if pol:
                if amp.shape[0] > 1:
                    amp_pol = rms_amp(np.sqrt(amp[1] ** 2 + amp[2] ** 2))
            amp = rms_amp(amp[0])

            if key == "cmb" or key == "cmb_lowl":
                freq_scaling = (np.ones(len(nu)) * Unit("uK_CMB")).to("uK_RJ", equivalencies=cmb_equivalencies(nu * u.GHz))
                seds[key][0, i, :] = 45 * freq_scaling  # TEMP
                seds[key][1, i, :] = 0.67 * freq_scaling  # POL
            else:
                freq_scaling = value.get_freq_scaling(nu * u.GHz, **specinds)
                k = 1 if freq_scaling.shape[0] > 1 else 0
                seds[key][0, i, :] = amp * freq_scaling[0]
                seds[key][1, i, :] = amp_pol * freq_scaling[k]

    return seds


def rms_amp(m):
    n = len(m) - np.sum(m.mask)
    m -= np.mean(m)
    return np.sqrt(np.sum(m**2) / n)



import inspect
import warnings
from typing import Dict, List, Literal, Optional, Type, Union

import healpy as hp
import numpy as np
from astropy.units import Quantity, Unit
from tqdm import tqdm

from cosmoglobe.h5._exceptions import (
    ChainComponentNotFoundError,
    ChainFormatError,
    ChainKeyError,
)
from cosmoglobe.h5.chain import Chain, ChainVersion
from cosmoglobe.sky._base_components import (
    DiffuseComponent,
    PointSourceComponent,
    SkyComponent,
)
from cosmoglobe.sky._chain_context import chain_context_registry
from cosmoglobe.sky.cosmoglobe import DEFAULT_COSMOGLOBE_MODEL, CosmoglobeModel

DEFAULT_SAMPLE = -1


def get_components_from_chain(
    chain: Union[str, Chain],
    nside: int,
    components: Optional[List[str]] = None,
    model: CosmoglobeModel = DEFAULT_COSMOGLOBE_MODEL,
    samples: Optional[Union[range, int, Literal["all"]]] = DEFAULT_SAMPLE,
    burn_in: Optional[int] = None,
) -> Dict[str, SkyComponent]:
    """Initialize and return a cosmoglobe sky model from a chainfile.

    Parameters
    ----------
    chain
        Path to a Cosmoglobe chainfile or a Chain object.
    nside
        Model HEALPIX map resolution parameter.
    components
        List of components to include in the model.
    model
        String representing which sky model to use. Defaults to the
        BeyondPlanck model.
    samples
        The sample number for which to extract the model. If the input
        is 'all', then the model will an average of all samples in the chain.
        Defaults to the last sample in the chain.
    burn_in
        Burn in sample for which all previous samples are disregarded.

    Returns
    -------
    initialized_components
        List containing initialized sky components

    """

    if not isinstance(chain, Chain):
        chain = Chain(chain, burn_in)

    if isinstance(samples, str):
        if samples.lower() == "all":
            samples = None
        else:
            raise ValueError("samples must be either 'all', an int, or a range")

    if chain.version is ChainVersion.OLD:
        raise ChainFormatError(
            "cannot initialize a sky model from a chain without a " "parameter group"
        )

    if ".astropy/cache" in str(chain.path):
        chain_name = "cached chainfile"
    else:
        chain_name = chain.path.name
    print(f"Initializing model from {chain_name}")
    if components is None:
        components = chain.components.copy()

        for component in chain.components:
            if component not in components:
                warnings.warn(
                    f"component {component} present in the Sky Model {model.version} "
                    "but not found in chain."
                )

    initialized_components: Dict[str, SkyComponent] = {}
    with tqdm(total=len(components), ncols=75) as progress_bar:
        padding = len(max(chain.components, key=len))
        for component in components:
            progress_bar.set_description(f"{component:<{padding}}")
            try:
                component_class = model[component]
            except KeyError:
                raise ChainComponentNotFoundError(
                    f"{component=!r} is not part in the Cosmoglobe Sky Model"
                )
            initialized_components[component] = get_comp_from_chain(
                chain, nside, component, component_class, samples
            )
            progress_bar.update()

    return initialized_components


def get_comp_from_chain(
    chain: Chain,
    nside: int,
    component_label: str,
    component_class: Type[SkyComponent],
    samples: Optional[Union[range, int]] = None,
) -> SkyComponent:
    """Initialize and return a sky component from a chainfile.

    Parameters
    ----------
    chain
        Chain object.
    nside
        Model HEALPIX map resolution parameter.
    component_label
        Name of the component.
    component_class
        Class representing the sky component.
    samples
        A range object, or an int representing the samples to use from the
        chain.

    Returns
    -------
        Initialized sky component object.
    """

    class_args = get_comp_signature(component_class)
    args: Dict[str, Quantity] = {}

    # Chain contexts are operations that we perform on the data in the chain
    # to fit it to the format required in the Cosmoglobe Sky Model. This includes,
    # renaming of variables (mappings), specifying astropy units, and reshaping
    # and/or converting maps constant over the sky to scalars.
    parameter_mappings = chain_context_registry.get_parameter_mappings(component_class)
    units = chain_context_registry.get_units(component_class)
    for arg in class_args:
        chain_arg = parameter_mappings.get(arg, arg)
        chain_params = chain.parameters[component_label]

        if chain_arg in chain_params:
            value = chain_params[chain_arg]
        else:
            is_alm = True if chain_arg == "amp" else False
            if is_alm:
                try:
                    value = chain.mean(
                        f"{component_label}/{chain_arg}_alm", samples=samples
                    )
                except ChainKeyError:
                    # Component is radio
                    try:
                        value = chain.mean(
                            f"{component_label}/{chain_arg}", samples=samples
                        )
                    except ChainKeyError:
                        value = chain.mean(
                            f"{component_label}/{chain_arg}_map", samples=samples
                        )
                    is_alm = False
            else:
                try:
                    value = chain.mean(
                        f"{component_label}/{chain_arg}", samples=samples
                    )
                except ChainKeyError:
                    value = chain.mean(
                        f"{component_label}/{chain_arg}_map", samples=samples
                    )

            nside_out = nside if nside is not None else chain_params["nside"]
            if is_alm:
                pol = True if arg == "amp" and value.shape[0] == 3 else False
                lmax = chain.get(f"{component_label}/{chain_arg}_lmax", samples=0)
                value = hp.alm2map(
                    value,
                    nside=nside_out,
                    lmax=lmax,
                    fwhm=(chain_params["fwhm"] * Unit("arcmin")).to("rad").value,
                    pol=pol,
                )
            else:
                try:
                    hp.npix2nside(value[0].size)
                    value = np.asarray(
                        [
                            hp.ud_grade(
                                value[IQU], nside_out=nside_out
                            )
                            for IQU in range(value.shape[0])
                        ]
                    )
                except ValueError:
                    pass


        args[arg] = Quantity(value, unit=units[arg] if arg in units else None)

    functions = chain_context_registry.get_functions(component_class)
    for function in functions:
        args.update(function(args))

    return component_class(**args)


def get_comp_signature(comp_class: Type[SkyComponent]) -> List[str]:
    """Extracts and returns the parameters to extract from the chain."""

    EXCLUDE = [
        "self",
        "freqs",
        "spectral_parameters",
        "_",
        "__",
    ]
    arguments: List[str] = []
    class_signature = inspect.signature(comp_class)
    arguments.extend(list(class_signature.parameters.keys()))
    if issubclass(comp_class, (DiffuseComponent, PointSourceComponent)):
        SED_signature = inspect.signature((comp_class.get_freq_scaling))
        arguments.extend(list(SED_signature.parameters.keys()))

    arguments = [arg for arg in arguments if arg not in EXCLUDE]

    return arguments



from typing import Dict, List, Protocol, Union

import numpy as np
from astropy.units import Quantity, Unit
from astropy.units.core import UnitBase
from scipy.interpolate import RectBivariateSpline

from cosmoglobe.sky._intensity_derivatives import get_intensity_derivative
from cosmoglobe.sky._units import cmb_equivalencies

# Dict mapping number of grid points to the interpolation dimensin of the
# bandpass integration.
N_INTERPOLATION_GRID = {
    1: 1000,
    2: 100,
}

def get_normalized_weights(
    freqs: Quantity,
    weights: Quantity,
    component_amp_unit: UnitBase,
) -> Quantity:
    """Normalizes a bandpass to units of unity under integration.

    Additionally, the bandpass is converted to the units of the amplitude
    map of the sky component used in the bandpass integration.

    Parameters
    ----------
    freqs
        Frequencies corresponding to bandpass weights.
    weights
        The weights of the bandpass.

    Returns
    -------
        Bandpass object.
    """

    # For pointsources, we need to divide the quantity by sr to make it
    # compatible with the cmb equivalencies.
    if component_amp_unit.is_equivalent("Jy"):
        component_amp_unit /= Unit("sr")

    weights = weights.to(component_amp_unit, equivalencies=cmb_equivalencies(freqs))
    weights /= np.trapz(weights, freqs)

    return weights


def get_bandpass_coefficient(
    freqs: Quantity,
    weights: Quantity,
    input_unit: Unit,
    output_unit: Unit,
) -> Quantity:
    """Returns the bandpass coefficient (Unit conversion factor for bandpass integrations).

    Parameters
    ----------
    freqs
        Frequencies corresponding to bandpass weights.
    weights
        The weights of the bandpass.
    input_unit
        The unit of the cosmoglobe data.
    ouput_unit
        The requested output unit of the simulated emission.

    Returns
    -------
    coefficient
        Bandpass coefficient representing the unitconversion between the
        input/ouput unit over the bandpass.
    """

    in_intensity_derivative = get_intensity_derivative(input_unit)
    out_intensity_derivative = get_intensity_derivative(output_unit)

    coefficient = np.trapz(weights * in_intensity_derivative(freqs), freqs) / np.trapz(
        weights * out_intensity_derivative(freqs), freqs
    )

    return coefficient


class FreqScalingFunc(Protocol):
    """Protocol defining the interface of the `get_freq_scaling` function of a SkyComponent."""

    def __call__(self, freqs: Quantity, **spectral_parameters: Quantity) -> Quantity:
        ...


class BandpassIntegration(Protocol):
    """Protocol defining the interface for a bandpass integration implementation."""

    def __call__(
        self,
        freqs: Quantity,
        weights: Quantity,
        freq_scaling_func: FreqScalingFunc,
        spectral_parameters: Dict[str, Quantity],
        *,
        interpolation_grid: Dict[str, Quantity],
    ) -> Union[float, List[float]]:
        """Computes the frequency scaling factor over bandpass integration.

        The bandpass integration is performed using the Commander mixing matrix
        formalism. Rather than computing the frequency scaling factor per pixel
        for each frequency in the bandpass, we grid out the spectral parameters
        of the component to a small range covering the spatial variations over
        IQU. We then compute the frequency scaling factor for each gridded
        parameter integrated over the bandpass and store these values. To
        estimate the bandpass integration factor, we then simply interpolate in
        these values.

        Parameters
        ----------
        freqs
            Frequencies corresponding to bandpass weights.
        weights
            The weights of the bandpass.
        freq_scaling_func
            Function that returns the SED scaling given frequencies and
            spectral parameters.
        spectral_parameters
            Dictionary containing the spectral parameters of the component we
            want to integrate over a bandpass.
        interpolation_grid
            Grid of spectral parameter values for the component. This grid
            used as interpolation values to compute the interpolated scaling
            factor.

        Returns
        -------
            Bandpass integration factor.
        """


def integrated_freq_scaling(
    freqs: Quantity,
    weights: Quantity,
    freq_scaling_func: FreqScalingFunc,
    spectral_parameters: Dict[str, Quantity],
    **_,
) -> float:
    """Bandpass integration 0D.

    The bandpass integration algorithm for a component with 0 spatially
    varying spectral parameters.
    """

    freq_scaling = freq_scaling_func(freqs, **spectral_parameters)
    scaling_factor = np.trapz(freq_scaling * weights, freqs)
    if np.ndim(scaling_factor) > 0:
        return np.expand_dims(scaling_factor, axis=1)

    return scaling_factor


def bandpass_interpolation_1d(
    freqs: Quantity,
    weights: Quantity,
    freq_scaling_func: FreqScalingFunc,
    spectral_parameters: Dict[str, Quantity],
    interpolation_grid: Dict[str, Quantity],
) -> List[float]:
    """Bandpass integration 1D

    The bandpass integration algorithm for a component with a single spatially
    varying spectral parameter."""

    # Grid dictionary only has one item
    ((key, spectral_parameter_grid),) = interpolation_grid.items()
    if spectral_parameters[key].shape[0] == 3:
        integrals = np.zeros((len(spectral_parameter_grid), 3))
    else:
        integrals = np.zeros((len(spectral_parameter_grid), 1))

    for idx, grid_point in enumerate(spectral_parameter_grid):
        scalar_params = {
            param: value for param, value in spectral_parameters.items() if param != key
        }
        freq_scaling = freq_scaling_func(freqs, **{key: grid_point}, **scalar_params)
        integrals[idx] = np.trapz(freq_scaling * weights, freqs)

    # We transpose the array to make it into row format similar to how
    # regular IQU maps are stored
    integrals = np.transpose(integrals)

    scaling = [
        np.interp(
            spectral_parameters[key][IQU].value,
            spectral_parameter_grid.value,
            integral,
        )
        for IQU, integral in enumerate(integrals)
    ]

    return scaling


def bandpass_interpolation_2d(
    freqs: Quantity,
    weights: Quantity,
    freq_scaling_func: FreqScalingFunc,
    spectral_parameters: Dict[str, Quantity],
    interpolation_grid: Dict[str, Quantity],
) -> List[float]:
    """Bandpass integration 2D

    The bandpass integration algorithm for a component with two spatially
    varying spectral parameters.
    """

    mesh_grid = {
        key: value
        for key, value in zip(
            interpolation_grid.keys(), np.meshgrid(*interpolation_grid.values())
        )
    }

    # Make n x n mesh grid for the spectral parameters
    n = len(list(interpolation_grid.values())[0])
    if any(
        spectral_param.shape[0] == 3 for spectral_param in spectral_parameters.values()
    ):
        integrals = np.zeros((n, n, 3))
    else:
        integrals = np.zeros((n, n, 1))

    for i in range(n):
        for j in range(n):
            grid_spectrals = {key: value[i, j] for key, value in mesh_grid.items()}
            freq_scaling = freq_scaling_func(freqs, **grid_spectrals)
            integrals[i, j] = np.trapz(freq_scaling * weights, freqs)
    integrals = np.transpose(integrals)

    scaling = []
    for IQU, integral in enumerate(integrals):
        f = RectBivariateSpline(*interpolation_grid.values(), integral)
        packed_spectrals = [
            spectral[IQU] if spectral.shape[0] == 3 else spectral[0]
            for spectral in spectral_parameters.values()
        ]
        scaling.append(f(*packed_spectrals, grid=False))

    return scaling


# Dict mapping number of the spatially varying spectral parameters to bandpass
# integration implementation.
BP_INTEGRATION_MAPPINGS: Dict[int, BandpassIntegration] = {
    0: integrated_freq_scaling,
    1: bandpass_interpolation_1d,
    2: bandpass_interpolation_2d,
}


def get_interpolation_grid(
    spectral_parameters: Dict[str, Quantity]
) -> Dict[str, Quantity]:
    """Returns a interpolation range.

    Computes the interpolation range of the spectral parameters of a
    sky component. We use a regular grid with n points for the range.

    Parameters
    ----------
    spectral_parameters
        Dictionary containing the spectral parameters of a given component.

    Returns
    -------
    interp_parameters
        Dictionary with a interpolation grid for each spatially varying
        spectral parameter.
    """

    dim = 0
    for spectral_parameter in spectral_parameters.values():
        if spectral_parameter.size > 3:
            dim += 1

    grid: Dict[str, Quantity] = {}
    if dim == 0:
        return grid

    try:
        n = N_INTERPOLATION_GRID[dim]
    except KeyError:
        raise NotImplementedError(
            "Bandpass integration for comps with more than two spectral "
            "parameters is not currently supported"
        )
    for key, value in spectral_parameters.items():
        if value.size > 3:
            grid_range = np.linspace(np.amin(value), np.amax(value), n)
            grid[key] = grid_range

    return grid


def get_bandpass_scaling(
    freqs: Quantity,
    weights: Quantity,
    freq_scaling_func: FreqScalingFunc,
    spectral_parameters: Dict[str, Quantity],
) -> Union[float, List[float]]:
    """Returns the appropriate Bandpass integration implementation given the spectral parameters."""

    grid = get_interpolation_grid(spectral_parameters)
    if (dim := len(grid)) not in BP_INTEGRATION_MAPPINGS:
        raise NotImplementedError(
            "Bandpass integration for comps with more than two spectral "
            "parameters is not currently supported"
        )

    return BP_INTEGRATION_MAPPINGS[dim](
        freqs=freqs,
        weights=weights,
        freq_scaling_func=freq_scaling_func,
        spectral_parameters=spectral_parameters,
        interpolation_grid=grid,
    )



import numpy as np
from astropy.units import Quantity, Unit, quantity_input

from cosmoglobe.sky._base_components import DiffuseComponent
from cosmoglobe.sky.components._labels import SkyComponentLabel


class LinearOpticallyThin(DiffuseComponent):
    r"""Class representing the free-free component in the Cosmoglobe Sky Model.

    Notes
    -----
    The free-free emission is defined using the convention in
    `BeyondPlanck (2020), Section 3.3.2
    <https://arxiv.org/pdf/2011.05609.pdf>`_;

    .. math::

        \boldsymbol{s}_\mathrm{RJ}^{\mathrm{ff}}(\nu) \propto
        \frac{g_{\mathrm{ff}}(T_\mathrm{e})}{\nu^2},

    where :math:`\nu` is the frequency for which we are simulating the
    sky emission, :math:`g_\mathrm{ff}` is the Gaunt factor, and
    :math:`T_\mathrm{e}` is the electron temperature.
    """

    label = SkyComponentLabel.FF
    freq_range = (0 * Unit("Hz"), 5 * Unit("THz"))

    def get_freq_scaling(self, freqs: Quantity, T_e: Quantity) -> Quantity:
        """See base class."""

        gaunt_factor_ratio = gaunt_factor(freqs, T_e) / gaunt_factor(self.freq_ref, T_e)
        scaling = (self.freq_ref / freqs) ** 2 * gaunt_factor_ratio

        return scaling


@quantity_input(freq="Hz", T_e="K")
def gaunt_factor(freq: Quantity, T_e: Quantity) -> Quantity:
    """Returns the Gaunt factor.

    Computes the gaunt factor for a given frequency and electron
    temperaturein SI units.

    Parameters
    ----------
    freq
        Frequency [Hz].
    T_e
        Electron temperature [K].

    Returns
    -------
    gaunt_factor
        Gaunt Factor.
    """

    # Avoiding overflow and underflow.
    T_e = T_e.astype(np.float64)
    T_e = (T_e.to("kK")).value / 10
    freq = (freq.to("GHz")).value

    gaunt_factor = np.log(
        np.exp(5.96 - (np.sqrt(3) / np.pi) * np.log(freq * T_e ** -1.5)) + np.e
    )

    return Quantity(gaunt_factor)



from functools import wraps

import h5py

from cosmoglobe.h5 import PARAMETER_GROUP_NAME
from cosmoglobe.h5._alms import unpack_alms_from_chain
from cosmoglobe.h5._exceptions import ChainKeyError, ChainSampleError


def validate_key(func):
    """Decotrator to check if the requested key exists in the chain."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        chain, key, *_ = args

        root = key.split("/")[0]
        if root.startswith(PARAMETER_GROUP_NAME) or root in chain.samples:
            path = key
        else:
            path = f"{chain.samples[0]}/{key}"

        with h5py.File(chain.path, "r") as file:
            try:
                file[path]
            except KeyError:
                raise ChainKeyError(f"{key=} not found in chain")
        return func(*args, **kwargs)

    return wrapper


def validate_samples(func):
    """Decotrator to check if the samples exist."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        chain, *_ = args
        samples = kwargs.get("samples")
        if samples is None:
            samples = chain.samples
        elif isinstance(samples, int):
            try:
                samples = [chain.samples[samples]]
            except IndexError:
                raise ChainSampleError(f"input sample {samples} is not in the chain")
        elif isinstance(samples, range):
            samples = list(samples)
        elif isinstance(samples, (tuple, list)):
            for sample in samples:
                try:
                    chain.samples[sample]
                except IndexError:
                    raise ChainSampleError(f"input sample {sample} is not in the chain")
            samples = list(samples)
        else:
            raise ChainSampleError("input samples must be an int or a range")

        if len(samples) > chain.nsamples:
            raise ChainSampleError(
                f"samples out of range with chain. chain only has {chain.nsamples} samples"
            )

        if all(isinstance(sample, int) for sample in samples):
            samples = chain._format_samples(samples)

        kwargs["samples"] = samples

        return func(*args, **kwargs)

    return wrapper


def unpack_alms(func):
    """Decotrator to that unpacks alms if they key is an alm."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        chain, key, *_ = args
        values = func(*args, **kwargs)

        if "alm" in key and kwargs.get("unpack") is not False:
            return unpack_alms_from_chain(chain, values, key)

        return values

    return wrapper



import matplotlib.pyplot as plt
import numpy as np

from .plottools import *
from cosmoglobe.h5.chain import Chain


def trace(
    input,
    dataset=None,
    sig=0,
    labels=None,
    showval=True,
    display_prec=2,
    burnin=0,
    xlabel=None,
    ylabel=None,
    nbins=None,
    cmap="tab10",
    figsize=(8, 3),
    darkmode=False,
    sub=None,
    hold=False,
    reuse_axes=False,
):
    """
    Make a trace plot of a quantity over gibbs samples

    Parameters
    ----------
    input : ndarray, fits file path or cosmoglobe model object
        Map data input given as numpy array either 1d or index given by 'sig'.
        Also supports fits-file path string or cosmoglobe model.
        If cosmoglobe object is passed such as 'model', specify comp or freq.
    dataset : str, optional
        if passing hdf5 file, specify dataset
        default: None
    sig : str or int, optional
        Specify which signal to plot if ndim>1.
        default: None
    labels : list of str, optional
        List of strings to use as labels for each of the components
        default: None
    showval : bool, optional
        display the mean and stddev of the distribution next to label
        default: True
    display_prec : int, optional
        sets precision of the mean and stddev of distribution.
        default: 2
    burnin : int, optional
        Number of burnin samples, used in calculating mean and stddev values
        default: 0
    xlabel : str, optional
        label x axis
        default: None
    ylabel : str, optional
        label y axis
        default: None
    nbins : int, optional,
        number of bins
        default: None
    cmap : str, optional
        Colormap (ex. sunburst, planck, jet). Both matplotliib and cmasher
        available as of now. Also supports qualitative plotly map, [ex.
        q-Plotly-4 (q for qualitative 4 for max color)] Sets planck as default.
        default: None
    figsize : touple, optional
        size of figure
        default: None
    darkmode : bool, optional
        turn all axis elements white for optimal dark visualization
        default: False
    sub : int, scalar or sequence, optional
        Use only a zone of the current figure (same syntax as subbplot).
        Default: None
    hold : bool, optional
        If True, replace the current Axes by a MollweideAxes.
        use this if you want to have multiple maps on the same
        figure.
        Default: False
    reuse_axes : bool, optional
        If True, reuse the current Axes (should be a MollweideAxes). This is
        useful if you want to overplot with a partially transparent colormap,
        such as for plotting a line integral convolution.
        Default: False
    """
    chain = None
    if isinstance(input, str):
        if input.endswith(".h5") and dataset is not None:
            chain = Chain(input)
            if ylabel == None:
                ylabel = dataset
            # component, *items = dataset.split("/")

            input = chain.get(dataset)

    # Make figure
    fig, ax = make_fig(
        figsize,
        None,
        hold,
        sub,
        reuse_axes,
        darkmode,
    )

    if input.ndim < 2:
        input.reshape(-1, 1, 1)
    elif chain is None and input.ndim == 2:
        input = input[:, np.newaxis, :]
    elif input.ndim == 2:
        input = input[:, :, np.newaxis]

    Nsamp, Nsig, Ncomp = input.shape

    cmap = load_cmap(cmap)
    positions = legend_positions(
        input[:, sig, :],
    )

    for i in range(Ncomp):
        plt.plot(
            input[:, sig, i],
            color=cmap(i),
            linewidth=2,
        )

        # Add the text to the right
        if labels is not None:
            hpos = Nsamp * 1.01
            plt.text(
                hpos,
                positions[i],
                rf"{labels[i]}",
                color=cmap(i),
                fontweight="normal",
            )

        if showval:
            mean = np.mean(input[burnin:, sig, i])
            std = np.std(input[burnin:, sig, i])
            label2 = rf"{mean:.{display_prec}f}$\pm${std:.{display_prec}f}"
            valpos = Nsamp * 1.01 if labels is None else Nsamp * 1.1
            plt.text(valpos, positions[i], label2, color=cmap(i), fontweight="normal")

    ax.set_xlim(right=Nsamp)
    ax.set_ylabel(ylabel)
    ax.set_xlabel(xlabel)

    # Tick and spine parameters
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.tick_params(axis="both", which="both", direction="in")
    plt.yticks(
        rotation=90,
        va="center",
    )
    plt.subplots_adjust(wspace=0, hspace=0.0, right=1)
    plt.tight_layout()



import numpy as np
import healpy as hp
import matplotlib.pyplot as plt
from .plottools import set_style, make_fig, get_percentile
from cosmoglobe.h5.chain import Chain
import scipy.stats as stats


def hist(
    input,
    *args,
    dataset=None,
    prior=None,
    xlabel=None,
    ylabel=None,
    field=0,
    darkmode=False,
    figsize=None,
    sub=None,
    hold=False,
    reuse_axes=False,
    **kwargs,
):
    """
    This function is a wrapper on the existing matplotlib histogram with custom styling.
    Possible to plot accompanying normalized prior distribution

    Parameters
    ----------
    input : array
        data input array
    prior : touple, optional
        mean and stddev of gaussian distribution overlay
        default: None
    dataset : str, optional
        dataset to plot if passing hdf5 file
        default: None
    xlabel : str, optional
        label x axis
        default: None
    ylabel : str, optional
        label y axis
        default: None
    field : int, optional
        which field to plot if .fits file is passed
        default: 0
    darkmode : bool, optional
        turn all axis elements white for optimal dark visualization
        default: False
    figsize : touple, optional
        size of figure
        default: None
    sub : int, scalar or sequence, optional
        Use only a zone of the current figure (same syntax as subplot).
        Default: None
    hold : bool, optional
        If True, replace the current Axes by a MollweideAxes.
        use this if you want to have multiple maps on the same
        figure.
        Default: False
    reuse_axes : bool, optional
        If True, reuse the current Axes (should be a MollweideAxes). This is
        useful if you want to overplot with a partially transparent colormap,
        such as for plotting a line integral convolution.
        Default: False

    """
    hp.mollview
    if isinstance(input, str):
        if input.endswith(".h5") and dataset is not None:
            chain = Chain(input)
            if ylabel == None:
                ylabel = dataset
            input = chain.get(dataset)
        elif input.endswith(".fits"):
            input = hp.read_map(input, field=field)
        else:
            raise ValueError("Input format not recognized")

    # Make figure
    fig, ax = make_fig(
        figsize,
        None,
        hold,
        sub,
        reuse_axes,
        darkmode,
    )

    # Use matplotlib histogram function with specific options
    n, bins, patches = plt.hist(input, *args, histtype="step", density=True, stacked=True, **kwargs)
    # Overlay a gaussian prior
    if prior is not None:
        dx = bins[1] - bins[0]
        x = np.linspace(bins[0], bins[-1], len(input))
        norm = sum(n) * dx
        Pprior = stats.norm.pdf(x, prior[0], prior[1])  # *norm
        plt.plot(
            x,
            Pprior * norm,
            linestyle=":",
            label=r"$\mathcal{N}(" + f"{prior[0]},{prior[1]}" + ")$",
        )

    # Tick and spine parameters
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.tick_params(axis="both", which="both", direction="in")
    plt.yticks(
        rotation=90,
        va="center",
    )
    ax.set_ylabel(ylabel)
    ax.set_xlabel(xlabel)

    return n, bins, patches
